#!/usr/bin/env python

from collections import OrderedDict
from datetime import datetime, timedelta
from multiprocessing import Pool
import glob
import os
import shutil

from shapely.geometry import Polygon
from scipy.constants import N_A
from scipy.interpolate import SmoothBivariateSpline
from xarray import DataArray, Dataset

import cartopy.crs as ccrs
import matplotlib.pyplot as plt
import numpy as np
import numba
import pandas
import scipy.integrate
import scipy.ndimage as ndi
import shapely

import skimage.morphology
import skimage.draw
import xarray

import doas
import dplume

from amrs.misc.chem import ppm_to_mass, kgs_to_Mtyr
import amrs.misc
import omi

from smartcarb import config as cfg
from smartcarb import misc
from smartcarb import map
import smartcarb.io


import scipy.optimize as sopt


from . import mass
from . import vis
from . import wind
import emiss.background
import emiss.misc
import emiss.io



# xlim and ylim for visualization of Gauss plume
FILENAME2LIM = {
    'Sentinel_7_CO2_2015011311_o0190_l270.nc': (-15,90,-15,90, 20),
    'Sentinel_7_CO2_2015012411_o0354_l270.nc': (-25,25,-15,110, 20),
}




def nox_to_no2(cm, model='Baechlin', inverse=False):
    """\
    Convert NOx concentrations (ug/m3) to NO2 concentrations (ug/m3) using
    empirical equation (During et al., 2011).
    """
    if model == 'Baechlin':
        A = 29.0
        B = 35.0
        C = 0.217
    elif model == 'Romberg':
        A = 103.0
        B = 130.0
        C = 0.005
    else:
        raise ValueError('Unknown model')

    if inverse:
        a = - (A + B*C - cm) / (2 * C)

        b = ((A + B*C - cm) / C)**2 + 4 * B * cm / C
        b = np.sqrt(b) / 2

        return a + b

    else:
        return  A * cm / (cm + B) + C * cm




def test_est_background(filename, tracer=None, ssp=None, nsat=6, nmax=100,
                        co2_ssp=None, use_ppm=True, is_mce=False, index=None):
    """
    Estimate XCO2 background of Berlin plume by normalized convolution using
    data outside plume. The estimate is compared to the true background
    known from model tracer.
    """

    bias = []
    scatter = []


    ts = pandas.read_csv(filename, parse_dates=True, index_col=0, sep='\t')

    if is_mce:
        ts, ts_std, ts_sizes = dplume.io.average_results(ts, add_size=True)

    if index is not None:
        ts = ts.loc[index]

    ts = misc.filter_dataset(ts, nsat=nsat)
    if use_ppm:
        bias = ts['CO2 background bias (ppm)']
        scatter = ts['CO2 background scatter (ppm)']
    else:
        bias = -1.0 * ts['CO2 background bias (Mt/yr)']
        scatter = ts['CO2 background scatter (Mt/yr)']


    bias = bias[np.isfinite(bias)]
    scatter = scatter[np.isfinite(scatter)]

    print('size:', len(bias), len(scatter))
    print('bias:')
    print('- mean:', np.mean(bias))
    print('- std:', np.std(bias))
    print('scatter:')
    print('- mean:', np.mean(scatter))
    print('- std:', np.std(scatter))

    # histograms
    fig = plt.figure(figsize=(7,3))

    ax = fig.add_subplot(1,2,1)
    if use_ppm:
        h = ax.hist(bias, bins=np.arange(-0.20, 0.21, 0.025))[0]
        ax.set_xlabel('mean bias (ppm)')
    else:
        h = ax.hist(bias, bins=np.arange(-20, 21, 2.5))[0]
        ax.set_xlabel('mean bias (Mt/yr)')
    ax.set_ylabel('number of plumes')
    ax.set_ylim(0, h.max()+1)
    ax.grid(True, ls='-', color='gray')

    ax = fig.add_subplot(1,2,2)
    if use_ppm:
        h = ax.hist(scatter, bins=np.arange(0.0, 0.41, 0.025))[0]
        ax.set_xlabel('standard deviation (ppm)')
    else:
        h = ax.hist(scatter, bins=np.arange(0, 41, 2.5))[0]
        ax.set_xlabel('standard deviation (Mt/yr)')
    #ax.set_ylabel('number of plumes')
    ax.set_ylim(0, h.max()+1)
    ax.grid(True, ls='-', color='gray')

    plt.tight_layout()

    #product = '%s_%s_and_%.1f_ppm' % (tracer, ssp, co2_ssp)
    #filename = 'Berlin_est_CO2_bg_hist_with_%s.png' % product
    #plt.savefig(filename)
    #plt.savefig(filename.replace('.png', '.pdf'))

    return bias, scatter




def convert_molec_to_mass(data, tracers, molar_mass=46.0055e-3):
    """
    Convert molec. cm-2 to kg m-2.
    """
    for tracer in tracers:
        values = data[tracer]
        attrs = values.attrs

        name = '%s_mass' % tracer
        data[name] = values * molar_mass / N_A * 1e4

        data[name].attrs.update(attrs)
        data[name].attrs['units'] = 'kg m-2'

    return data




def convert_ppm_to_mass(data, tracers, gas='CO2'):
    for tracer in tracers:
        values = data[tracer]
        attrs = values.attrs

        name = '%s_mass' % tracer
        data[name] = ppm_to_mass(values, data.psurf, gas)

        if 'noise_level' in attrs and not isinstance(attrs['noise_level'], str):
            level = ppm_to_mass(attrs['noise_level'], np.nanmean(data.psurf), gas)
            attrs['noise_level'] = level

        data[name].attrs.update(attrs)
        data[name].attrs['units'] = 'kg m-2'

    return data


@numba.jit(nopython=True)
def distance(q,p):
    """Distance between two points."""
    return np.sqrt((q[0] - p[0])**2 + (q[1] - p[1])**2)


def find_min_distance(p,curve):

    def f(t):
        q = curve(t=t)
        return distance(q,p)

    t0 = curve.get_parameter(x=p[0], y=p[1])

    res = sopt.minimize(f, x0=t0,
            method='CG',
            options={'maxiter': 100},
    )
    q = curve(t=float(res.x))

    return p, q, res.fun, res




class Poly2D:
    def __init__(self, x, y, w, degree=2, x_o=0.0, y_o=0.0,
                 x0=0.0, y0=0.0, force_source=True):
        """\
        A 2D curve fitted on the point cloud given by x and y.

        Parameters
        ----------
        x,y,w: x,y coords and weights used for fitting the data

        degree: degrees if the two polynomials x(t) and y(t)

        x_o, y_o: location of source (will be added to x,y and given
                  high weight such that curves goes through source if
                  force_origin is True)

        x0, y0: origin of coordinate system
        """
        self.x = x
        self.y = y
        self.w = w
        self.degree = degree
        self.force_source = force_source

        self.x_o = x_o
        self.y_o = y_o
        self.t_o = np.nan

        self.x0 = x0
        self.y0 = y0

        self.c = np.zeros(2 * (degree + 1))

        self._fit()

        # arc length to origin
        self.t_o = self.get_parameter(self.x_o, self.y_o)

        # 
        self.tmin = 0.0
        self.tmax = np.max(self.get_parameter(self.x, self.y))
        self.interpolator = None


    def to_file(self, filename, group=None, mode='a'):
        """
        Convert Poly2D to netCDF file.
        """
        data = xarray.Dataset()

        for name in ['x', 'y', 'w', 'degree', 'x_o', 'y_o', 'x0', 'y0']:

            value = getattr(self, name)

            if np.ndim(value) == 0:
                data.attrs[name] = value

            else:
                data[name] = xarray.DataArray(value, dims='pixels')

        try:
            data.to_netcdf(filename, mode=mode, group=group)
        except FileNotFoundError:
            data.to_netcdf(filename, mode='w', group=group)
            
        return data


    @classmethod
    def from_file(cls, filename, group=None):
        """
        Read curve from netCDF file.
        """
        with xarray.open_dataset(filename, group=group) as data:
            args = dict((key, np.array(data[key])) for key in data)
            args.update(data.attrs)

        return cls(**args)


    def _fit(self):

        def objective(c, w, x, y):
            xt,yt = self._compute_curve(c,x,y)
            return np.concatenate([
                w * (xt - x), w * (yt - y)
            ])

        # add origin
        if self.force_source:
            x = np.append(self.x, self.x_o)
            y = np.append(self.y, self.y_o)
            w = np.append(self.w, 1000.)
        else:
            x = np.append(self.x, self.x0) # force origin of coords
            y = np.append(self.y, self.y0)
            w = np.append(self.w, 1000.)

        # angle around origin (not used)
        #phi = np.arctan2(y - y0, x - x0)

        # curve fit
        res = sopt.leastsq(objective, x0=self.c, args=(w,x,y), full_output=True)
        self.c = res[0]
        self.cov_x = res[1] # TODO
        ierr = res[4]

        if ierr not in [1,2,3,4]:
            print('least square failed with error code: ', res)


    def arc2parameter(self, arc):

        if self.interpolator is None:
            ta = np.arange(self.tmin - 50e3, 2 * self.tmax, 500.0)
            la = [arc_length_of_2d_curve(self, self.t_o, v) for v in ta]
            self.interpolator = scipy.interpolate.interp1d(la, ta)

        return self.interpolator(arc)


    def get_parameter(self, x, y):
        return np.sqrt((x - self.x0)**2 + (y - self.y0)**2)


    def compute_tangent(self, t0, norm=False):
        v = np.array(self(t=t0, m=1))
        if norm:
            v /= np.linalg.norm(v, axis=0)
        return v
    
    
    def compute_angle(self, t=None):
        """
        Compute tangent angle for curve.
        """
        if t is None:
            t = self.t_o
            
        u, v = self.compute_tangent(t)
        return np.rad2deg(np.arctan2(u,v))


    def get_coefficients(self, c=None, m=0):

        if c is None:
            c = self.c

        k = c.size // 2
        cx = c[:k]
        cy = c[k:]

        if m != 0:
            cx = np.polyder(cx, m)
            cy = np.polyder(cy, m)

        return cx, cy


    def compute_normal(self, t0, x=None, y=None, t=None):

        x0, y0 = self(t=t0)

        # tangent vector
        v = self.compute_tangent(t0, norm=True)

        # rotate 90 degree
        n = np.dot([[0,-1],[1,0]], v)

        cx = np.array([-n[0], x0])
        cy = np.array([-n[1], y0])

        if t is None:
            s = np.sign(v[0] * (y - y0) - v[1] * (x - x0))
            t = s * np.sqrt((x-x0)**2 + (y-y0)**2)

        return compute_poly_curve(cx, cy, t)
    

    def _compute_curve(self, c, x=None, y=None, t=None, m=0):

        if t is None:
            t = self.get_parameter(x,y)
        else:
            if x is not None or y is not None:
                print('Warning: `x` and `y` will be ignored as `t` was given.')

        cx, cy = self.get_coefficients(c=c, m=m)

        return compute_poly_curve(cx, cy, t)


    def __call__(self, x=None, y=None, t=None, m=0):
        return self._compute_curve(self.c, x, y, t, m)


def compute_poly_curve(cx,cy,t):
    """
    Compute poly curve.
    """
    return np.polyval(cx,t), np.polyval(cy,t)



def gauss2d(p, x, y, u, x0=0.0, y0=0.0):
    """\
    Computes Gaussian plume (units: kg/m2).

    p: state vector (Q, K, BG)
    - Q:  emissions in kg/s
    - K:  eddy diffusion coefficient in m2/s
    - BG: background field in kg/m2

    x: distance from origin in m
    y: distance from center line in m
    u: wind speed in m/s

    x0: x-coordinate of origin
    y0: y-coordinate of origin

    TODO:
    - decay time

    """
    # state vector
    Q = float(p[0])    # emissions (kg/s)
    K = float(p[1])    # eddy diffusion coefficient (m2 / s)

    if len(p) == 3:
        BG = float(p[2])   # background field (kg/m2)
    else:
        BG = 0.0

    # dispersion along the plume
    down = x > x0
    sigma = np.sqrt(2.0 * K * (x[down] - x0) / u)

    # compute Gaussian plume
    c = Q / (np.sqrt(2.0 * np.pi) * sigma * u) * np.exp(-0.5 * (y[down]-y0)**2 / sigma**2)

    plume = np.full(x.shape, BG)
    plume[down] += c

    return plume




def find_all_roots(hh, dd, d, h_max=5000.0):
    """\
    Find all heights (hh), below hh_max where wind direction (dd) equals
    given wind direction (d).
    """

    founds = []

    for i in range(hh.size-1):
        if dd[i] < dd[i+1]:
            dmin, dmax = dd[i], dd[i+1]
            hmin, hmax = hh[i], hh[i+1]
        else:
            dmin, dmax = dd[i+1], dd[i]
            hmin, hmax = hh[i+1], hh[i]

        if hmin > h_max:
            continue

        if dmin <= d < dmax:
            d0 = np.interp(d, [dmin, dmax], [hmin, hmax])
            founds.append(d0)

    return founds




def arc_length_of_poly(f, c, a, b):

    x = np.linspace(a,b)

    s = np.sqrt( 1 + np.polyval(np.polyder(c), x)**2 )

    return scipy.integrate.simps(s,x)


def arc_length_of_2d_curve(curve, a, b):

    t = np.linspace(a,b, 50001)
    xt, yt = curve(t=t, m=1)

    v = np.sqrt( xt**2 + yt**2)

    l = scipy.integrate.simps(v,t)
    return l




def compute_arc_length_and_distance(points, curve):
    """\
    Given n points and polynomial function on [x_o, x_max], compute
    arc length along curve and distance from curve.
    """
    dxs = []
    dys = []

    t_o = curve.get_parameter(curve.x_o, curve.y_o)
    args = ([point, curve] for point in points)

    with Pool(16) as pool:

        for (px,py), (qx,qy), dy, res in pool.starmap(find_min_distance, args):

            if np.isnan(dy):
                dx = np.nan
                print([px,py])
            else:
                # tanget vector
                t = curve.get_parameter(qx, qy)
                v = curve.compute_tangent(t, norm=True)

                # sign of distance (negative left of curve)
                sign = np.sign(np.cross([px-qx,py-qy],v))
                dy *= sign

                # arc length
                dx = arc_length_of_2d_curve(curve, t_o, t)

            dxs.append(dx)
            dys.append(dy)

    return np.array(dxs), np.array(dys)



def gauss2dmean(p,xc,yc,wind,delta=100.0):
    """
    Oversample field within satellite pixel and average.
    """
    n, _ = xc.shape
    zc = np.full(n, np.nan)

    for i in range(n):

        #px = shapely.geometry.Polygon(np.transpose([xc[i],yc[i]]))
        #if not px.is_valid:
        #    print('Error: Pixel #%d is not valid.' % i)
        #    continue

        # compute lower and upper boundary of pixel (as integers)
        xmin = np.ceil(xc[i].min())
        xmax = np.floor(xc[i].max())
        m = round((xmax - xmin) / delta + 1)
        xt = np.linspace(xmin,xmax+1, m)

        start = np.argmin(xc[i])
        lower = omi.cgrate.compute_line(xt, xc[i], yc[i], start, +1)
        upper = omi.cgrate.compute_line(xt, xc[i], yc[i], start, -1)

        if np.all(lower > upper):
            lower, upper = upper, lower

        # stack coordinates
        xx = []
        yy = []

        for j in range(xt.size):
            m = round((upper[j] - lower[j]) / delta + 1)

            if m < 0:
                print('Value negative:', i, lower[j], upper[j])
                continue

            y = np.linspace(lower[j], upper[j], m)
            x = np.full(y.shape, xt[j])

            xx.append(x)
            yy.append(y)

        xx = np.concatenate(xx)
        yy = np.concatenate(yy)

        zc[i] = np.mean(gauss2d(p,xx,yy,wind))

    return zc





def flux_inversion(p0, xp, yp, zp, wind, ssp, use_mean=True):

    args = (xp, yp, wind, zp)

    if use_mean:
        def model(p):
            return (gauss2dmean(p,xp,yp,wind) - zp) / ssp
    else:
        def model(p):
            return (gauss2d(p,xp,yp,wind) - zp) / ssp


    return scipy.optimize.least_squares(model, p0, method='lm')



def create_center_line(data, source, degree=2, force_source=False,
                       use_true_plume=False, tracer=None, source_distance=50e3):
    """
    source_distance: distance of the origin from the source in reverse
                     direction of the detected plume (default: 50e3 meters)
    """

    if use_true_plume:
        plume = data['%s_true_plume' % tracer].values
    else:
        plume = data['detected_plume'].values

    # Compute weights
    if data.attrs['tracer'] == 'NO2':
        weights = data['local_NO2_mean'] - data['NO2_estimated_background']
        weights = weights.values / 1e15

        weights[data.plume_area & ~plume] = 0.2
        weights[weights < 0.2] = 0.2
    else:
        weights = data['local_CO2_mean'] - data['CO2_estimated_background']
        weights = weights.values

        weights[data.plume_area & ~plume] = 0.05
        weights[weights < 0.05] = 0.05


    # get location of plume and XCO2 values (plus wind etc.)
    is_plume = compute_plume_area(data, distance=5e3, source=source, add_to_dataset=False)
    #is_plume = data.plume_area.values # FIXME

    x = np.asarray(data.x)[is_plume]
    y = np.asarray(data.y)[is_plume]
    w = np.asarray(weights)[is_plume]
    
    # location of the source
    lon_o, lat_o, _ = misc.get_source_origin(source, data=data)
    x_o, y_o = misc.wgs2epsg(lon_o, lat_o, epsg=data.x.attrs['epsg'], inverse=False)

    # origin of center line (50 km upstream of source)
    x0 = np.mean( data.x.values[plume] )
    y0 = np.mean( data.y.values[plume] )
    delta = np.sqrt((x0 - x_o)**2 + (y0 - y_o)**2)
    
    x1 = x_o - (x0 - x_o) / delta * source_distance
    y1 = y_o - (y0 - y_o) / delta * source_distance
    
    # fit 2d curve
    curve = Poly2D(x, y, w, degree, x_o=x_o, y_o=y_o, x0=x1, y0=y1,
                   force_source=force_source)

    return curve


def compute_xy_coords(data, epsg=cfg.EPSG):
    """
    Compute xy coordinates (in meters) for pixel centers and corners using EPSG
    number.
    """
    data['x'], data['y'] = misc.xy2wgs(data.lon, data.lat,
                                       epsg=epsg, inverse=True)

    if 'xc' in data and 'yc' in data:
        data['xc'], data['yc'] = misc.xy2wgs(data.lonc, data.latc,
                                             epsg=epsg, inverse=True)

    return data


def compute_plume_area(data, distance=50e3, source=None, add_to_dataset=True):
    """
    distance in meters, origin used to draw circle around
    """
    n = int(distance // 1e3 + 1)
    r = int(distance // 2e3 + 1) 

    shape = (n, n)
    kernel = np.zeros(shape, dtype=bool)

    rr,cc = skimage.draw.circle(r-1,r-1,r)
    kernel[rr,cc] = True

    # set pixels within distance from detected pixels to True
    if 'source' in data.dims:
        detected_plume = data['detected_plume'].sel(source=source)
    else:
        detected_plume = data['detected_plume']
        
    area = skimage.morphology.dilation(np.array(detected_plume), kernel)
    
    if add_to_dataset:

        # set area 50 km around source to True
        if source is not None:
            this = data.sel(source=source)
            lon_o = float(this['lon_o'])
            lat_o = float(this['lat_o'])
            x_o, y_o = misc.wgs2epsg(lon_o, lat_o, data.x.attrs['epsg'])
            area[ np.sqrt((data.x - x_o)**2 + (data.y - y_o)**2) < distance] = True

        if 'plume_area' not in data:
            data['plume_area'] = xarray.full_like(data.detected_plume, False)

        if 'source' in data.dims:

            # use index, because sel method might fail according to documentation
            index = int(np.argmax(data.source == source).values)
            data['plume_area'][:,:,index] = area

        else:
            data['plume_area'][:] = area

        return data

    else:
        return area



def compute_plume_coords(data, curve, source, do_corners=False):
    """
    Compute plume coordinates for center and corner pixels within plume area.
    """
    if 'source' in data:
        index = int(np.argmax(data.source == source).values)
        source_data = data.sel(source=source)
    else:
        index = 0
        source_data = data

    # center pixels in plume coords
   
    if 'xp' not in data or 'yp' not in data:
        if 'source' in data:
            shape = data.nobs.size, data.nrows.size, data.source.size
        else:
            shape = data.nobs.size, data.nrows.size, 1
        data['xp'] = xarray.DataArray(np.full(shape, np.nan), dims=('nobs', 'nrows', 'source'))
        data['yp'] = xarray.DataArray(np.full(shape, np.nan), dims=('nobs', 'nrows', 'source'))
        data.xp.attrs['long name'] = 'along plume coordiate'
        data.yp.attrs['long name'] = 'across plume coordinate'

    data['xp'][:,:,index], data['yp'][:,:,index] = emiss.misc.compute_plume_coordinates(source_data,
                                                                                        curve,
                                                                                        which='centers')

    # compute for all ground pixels distance to center line
    # pixel corners
    if do_corners:
        if 'xcp' not in data or 'ycp' not in data:
            shape = data.nobs.size, data.nrows.size, data.ncorners.size, data.source.size
            data['xcp'] = xarray.DataArray(np.full(shape, np.nan), dims=('nobs', 'nrows', 'ncorners', 'source'))
            data['ycp'] = xarray.DataArray(np.full(shape, np.nan), dims=('nobs', 'nrows', 'ncorners', 'source'))
            data.xcp.attrs['long name'] = 'along plume corner coordinates'
            data.ycp.attrs['long name'] = 'across plume corner coordinates'

        data['xcp'][:,:,:,index], data['ycp'][:,:,:,index] = emiss.misc.compute_plume_coordinates(
            source_data, curve, which='corners')

        # check if pixels still valid polygons (can happen if curve is
        # over- or undershooting
        for i,j in np.ndindex(source_data.plume_area.shape):

            if not source_data.plume_area[i,j]:
                continue

            coords = np.transpose([data['xcp'][i,j,:,index],data['ycp'][i,j,:,index]])
            px = shapely.geometry.Polygon(coords)

            if np.any(np.isnan(coords)) or not px.is_valid \
               or np.abs(px.area - px.convex_hull.area) > 1.0:
                print('Error: Mask invalid pixel corners (%d,%d).' % (i,j))
                data['xcp'][i,j,:,index] = np.nan
                data['ycp'][i,j,:,index] = np.nan


    # update plume area by masking nans
    invalids = np.isnan(data.xp[:,:,index]) | np.isnan(data.yp[:,:,index])
    
    if do_corners:
        invalids |= np.any(np.isnan(data.xcp[:,:,index]), axis=2)
        invalids |= np.any(np.isnan(data.ycp[:,:,index]), axis=2)

    if 'source' in data:
        area = np.array(data.plume_area.sel(source=source))
        area[invalids] = False
        data['plume_area'][:,:,index] = area
    else:
        area = np.array(data.plume_area)
        area[invalids] = False
        data['plume_area'][:,:] = area
        

    return data


def gnfra(bounds=None):
    """
    Weighting-factors for GNFR A emission profile
    """
    lower_bounds = np.array([0,  4, 30, 60,  90, 125, 170, 310, 470, 710])
    upper_bounds = np.array([4, 30, 60, 90, 125, 170, 310, 470, 710, 990])
    values = np.array([0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.08, 0.46, 0.29, 0.17])
    
    h = np.arange(0, 1000.0, 0.01)
    p = np.zeros_like(h)
    
    for v, low, up in zip(values, lower_bounds, upper_bounds):
        p[(low <= h) & (h < up)] = v
        
    
    # integrate
    p_bounds = np.zeros(bounds.shape[1])
    for i, (low, up) in enumerate(bounds.T):
        p_bounds[i] = p[(low <= h) & (h < up)].sum()
        
    return p_bounds / p_bounds.sum() 


def estimate_wind_speed(layers, uwind, vwind, curve=None, d_sat=None,
                        hmin=0.0, hmax=1e3, show=False, use_mean=False, gnfr='A'):
    """
    Estimate wind speed at source origin by matching plume propagation
    direction and wind profiles.
    """
    if layers[0] > layers[1]:
        layers = layers[::-1]
        uwind = uwind[::-1]
        vwind = vwind[::-1]

    layers = layers - layers[0]
    levels = 0.5 * (layers[1:] + layers[:-1])
    
    #mask = (hmin <= levels) & (levels < hmax)
    #levels = levels[mask]
    #uwind = uwind[mask]
    #vwind = vwind[mask]
    #wind_mean = np.sqrt(uwind**2 + vwind**2).mean()
    #direction_mean = misc.wind_direction(uwind.mean() ,vwind.mean())
    
    
    if gnfr is None:
        z2 = np.linspace(hmin, hmax, 5001)
        u2 = np.interp(z2, levels, uwind)
        v2 = np.interp(z2, levels, vwind)

        wind_mean = np.sqrt(u2**2 + v2**2).mean()
        direction_mean = misc.wind_direction(u2.mean(), v2.mean())
        
    else:
        wind = np.sqrt(uwind**2 + vwind**2)
        bounds = np.array([layers[:-1], layers[1:]])
        profile = gnfra(bounds)

        uwind_eff = float(sum(profile * uwind))
        vwind_eff = float(sum(profile * vwind))
        
        wind_mean = np.sqrt(uwind_eff**2 + vwind_eff**2)
        direction_mean = misc.wind_direction(uwind_eff, vwind_eff)
        
        

    if use_mean:
        params = {
            'mean wind speed (m/s)': float(wind_mean),
            'mean wind direction (deg)': float(direction_mean),
            'est. wind direction (m/s)': np.nan,
            'est. wind direction (m/s)': np.nan,
            'plume height (m)': np.nan,
        }
        return params

    raise NotImplementedError



    # wind direction (satellite)
    if curve is not None:
        xt,yt = curve(curve.x_o, curve.y_o, m=1)
        d_sat = misc.wind_direction(xt,yt)

    delta = misc.anglediff(direction, d_sat, is_degree=True)

    # find roots (if any)
    roots = misc.find_all_roots(levels, delta)

    if len(roots) == 0:
        # TODO: find minimum instead of root
        print('Warning: Wind and plume direction do not match exactly '
              'and thus using mean wind.')
        plume_height = np.nan
        wind_plume = np.mean(wind)
    else:
        # TODO: check if other heights are okay too
        # TODO: consider vertical average
        plume_height = roots[0]
        wind_plume = np.interp(plume_height, levels, wind)

    params = {
        'mean wind (m/s)': float(wind_mean),
        'est. wind (m/s)': float(wind_plume),
        'est. wind direction (deg)': float(d_sat),
        'plume height (m)': float(plume_height),
    }

    if show:
        fig = plt.figure()
        ax = fig.add_subplot(121)
        ax.plot(wind, levels)
        ax.set_xlim(xmin=0)
        ax.set_ylim(hmin, hmax)
        ax.set_ylabel('height (m)')
        ax.set_xlabel('wind speed (m/s)')
        ax.grid(True)

        ax = fig.add_subplot(122)
        ax.plot(delta, levels)
        ax.set_xlim(-50,50)
        ax.set_ylim(hmin, hmax)
        #ax.set_ylabel('height (m)')
        ax.set_xlabel('difference (degrees)')
        ax.grid(True)

        plt.tight_layout()


    return params



def estimate_plume_height(levels, wind, d_sat, hmin=0, hmax=2000.0):

    mask = (hmin <= hh) & (hh < hmax)

    # brute-force match wind direction between plume and model
    k_o = np.argmin(misc.anglediff(dd, d_sat)**2)
    wind = wind[k_o]

    # fine-tuning
    hh2 = np.linspace(
        hh[max(k_o-1, 0)],
        hh[min(k_o+1, hh.size-1)],
        51
    )
    uu2 = np.interp(hh2, hh, uu)
    vv2 = np.interp(hh2, hh, vv)
    dd2 = (270.0 - np.rad2deg(np.arctan2(vv2,uu2))) % 360

    k2 = np.argmin(misc.anglediff(dd2, d_sat)**2)
    wind = np.sqrt(uu2[k2]**2 + vv2[k2]**2)

    if hh2[k2] > hmax:
        print('Warning: Plume height %d m > %d m (hmax) ' % (hh2[k2], hmax))
        wind = wind_2km

    return wind, hh2[k2], wind_2km, dd2[k2] - d_sat


def draw_situation(field_vector, plume_vector, curve, ax=None, ms=20,
                   vmin=None, vmax=None):

    t = curve.get_parameter(field_vector.x, field_vector.y)
    t = np.linspace(t.min(), t.max())

    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111, aspect='equal')
    else:
        fig = plt.gcf()

    ax.plot(*curve(t=t), 'k-')

    c = ax.scatter(field_vector.x, field_vector.y, c=field_vector.a,
                   vmin=vmin, vmax=vmax, s=ms)

    ax.scatter(plume_vector.x, plume_vector.y, edgecolors='#222222',
               c='none', s=ms)

    ax.grid(True)
    ax.set_xlabel('x (km)')
    ax.set_ylabel('y (km)')

    fig.subplots_adjust(right=0.8)
    cax = fig.add_axes([0.82, 0.15, 0.03, 0.7])
    cb = fig.colorbar(c, cax=cax, extend='both')
    cb.set_label('XCO$_2$ (ppm)')



def visualize(x_c, y_c, field, x, dx, dy, wind, plume, filename):

    v = round(np.mean(field.a))
    s = max(1, round(np.std(field.a)))
    vmin, vmax = v-s, v+s

    fig1 = plt.figure()
    fig2 = plt.figure()

    ax1 = fig1.add_subplot(111, aspect='equal')
    ax2 = fig2.add_subplot(111, aspect='equal')


    ax1.set_ylabel('y (km)')
    ax1.set_title('(a) Observations')
    ax2.set_title('(b) Fit')

    try:
        xmin,xmax,ymin,ymax,ms = FILENAME2LIM[filename]
    except KeyError:
        xmin=xmax=ymin=ymax=None
        ms = None

    c = ax1.scatter(field.x, field.y, c=field.a, vmin=vmin, vmax=vmax, s=ms)
    c = ax2.scatter(field.x, field.y, c=gauss2d(x,dx,dy, u=wind),
                    vmin=vmin, vmax=vmax, s=ms)
    ax1.scatter(plume.x, plume.y, edgecolors='#222222', c='none', s=ms)


    for ax in [ax1, ax2]:

        ax.set_xlim(xmin,xmax)
        ax.set_ylim(ymin,ymax)

        ax.set_xlabel('x (km)')
        ax.plot(x_c, y_c, 'k-')

        # draw arrow
        d1,d2 = x_c[-1] - x_c[-2], y_c[-1] - y_c[-2]
        d1 = d1 / np.sqrt(d1**2 + d2**2)
        d2 = d2 / np.sqrt(d1**2 + d2**2)
        ax.arrow(x_c[-1], y_c[-1], d1, d2, width=0, lw=1, head_width=2,
                head_length=2, fc='k', ec='k')

        ax.grid(True)

    #label = io.filename2label(filename)
    #text = ', '.join([
    #    'Q = %.1f Mt/y' % x[0],
    #    'K = %d m$^2$/s' % x[1],
    #    'x$_{BG}$ = %d ppm' % x[2],
    #    'u = %.1f m/s' % wind
    #])
    #fig.suptitle('Gaussian plume fit for Janschwalde: %s\n(%s)' % (label, text), fontsize=10, y=0.99)


    for fig in [fig1, fig2]:
        fig.subplots_adjust(right=0.8)
        cax = fig.add_axes([0.82, 0.15, 0.03, 0.7])
        cb = fig.colorbar(c, cax=cax, extend='both')
        cb.set_label('XCO$_2$ (ppm)')





def fill_gaps(values, gaps, area=Ellipsis, sigma=2.0):
    """
    Fill gaps in area using normalized convolution with a Gaussian filter.

    values: array with gaps
    gaps: boolean array (true if gap)
    area: boolean array (true where gaps shall be filled)

    """
    kernel = emiss.background.create_gaussian_kernel(sigma)

    values = np.array(values)

    valids = np.isfinite(values)
    values[gaps] = amrs.misc.normalized_convolution(values, kernel,
                                                    mask=~valids)[gaps]
    # set values outside area to nan
    if area is not Ellipsis:
        values[~area] = np.nan

    if np.any(np.isnan(values[area])):
        print('Warning: Plume still has data gaps after gap filling.')

    return values



def plot_plume(data, curve, tracer='CO2_signal', origin='Berlin',
               add_cross_sections=True, contour=None, vmin=None, vmax=None,
               zoom_area=None):


    if isinstance(tracer, str):
        tracer = data[tracer]

    fig = map.make_level2_map(data.lonc, data.latc, tracer,
                              vmin=vmin, vmax=vmax,
                              label='XCO$_2$ signal (ppm)',
                              do_zoom=True, zoom_area=zoom_area)
    map.add_scatter(fig.axes[0], data.lon, data.lat, ms=4, marker='+',
                    slice=data.detected_plume, label='detected plume')
    map.add_scatter(fig.axes[0], data.lon, data.lat, ms=1, marker='o',
                    slice=data.plume_area & ~data.detected_plume, label='plume area')

    if "CO2_true_plume" in data:
        amrs.vis.draw_contours(fig.axes[0], data.lon, data.lat,
                               data.CO2_true_plume, label='CO$_2$ plume')

    t = curve.t_o + np.linspace(-4e3,
                                np.nanmax(np.array(data.xp)[data.plume_area]))

    color = 'y' if contour is None else 'k'
    map.add_curve(fig.axes[0], *curve(t=t), origin=origin, color=color)

    if add_cross_sections:
        if origin == 'Berlin':
            distances = np.arange(0, 1e3 + data.xp.max(), 10e3)
            t = np.linspace(-30e3, 30e3)
        else:
            distances = np.arange(0, 1e3 + data.xp.max(), 2e3)
            t = np.linspace(-5e3, 5e3)

        for t0 in distances: # every five km
            t0 += curve.t_o
            x,y = curve.compute_normal(t0=t0, t=t)
            map.add_curve(fig.axes[0], x, y, origin=origin, color='y')

    map.add_city_labels(fig.axes[0])

    if contour is not None:
        fig.axes[0].contour(data.lon, data.lat, contour, vmin=vmin, vmax=vmax,
                            levels=np.arange(vmin, vmax+0.1, 0.5),
                            transform=ccrs.PlateCarree())





def draw_filled_gaps(data):
    # XCO2 signal with filled gaps
    fig = map.make_level2_map(data.lonc, data.latc, data.CO2_signal,
                              vmin=-2, vmax=2, do_zoom=True,
                              domain='smartcarb',
                              label='XCO$_2$ signal (ppm)')

    lines = map.add_scatter(fig.axes[0], data.lon, data.lat, ms=0.7,
                    slice=data.detected_plume, label='detected plume')
    lines += amrs.vis.draw_contours(fig.axes[0], data.lon, data.lat,
                           data.CO2_true_plume, label='CO$_2$ plume')
    map.add_city_labels(fig.axes[0])
    fig.canvas.set_window_title('XCO2 signal (gaps filled)')

    fig.legend(lines, [line.get_label() for line in lines],
               loc=(0.025, 0.89), numpoints=1, markerscale=1.5,
               fontsize='small')

    # XCO2 signal with gaps
    fig = map.make_level2_map(data.lonc, data.latc, data.CO2 - data.CO2_background,
                              vmin=-2, vmax=2, do_zoom=True,
                              domain='smartcarb',
                              label='XCO$_2$ signal (ppm)')

    lines = map.add_scatter(fig.axes[0], data.lon, data.lat, ms=0.7,
                    slice=data.detected_plume, label='detected plume')
    lines += amrs.vis.draw_contours(fig.axes[0], data.lon, data.lat,
                           data.CO2_true_plume, label='CO$_2$ plume')
    map.add_city_labels(fig.axes)
    fig.canvas.set_window_title('XCO2 signal (with data gaps)')

    fig.legend(lines, [line.get_label() for line in lines],
               loc=(0.025, 0.89), numpoints=1, markerscale=1.5,
               fontsize='small')

    return lines


def all_gauss_1d(tracer, ssp, co2_ssp, origin='Berlin'):

    #filenames = sorted(fn[:-8] for fn in glob.glob('*.nc'))
    filenames = np.loadtxt('%sPlumes.dat' % origin, dtype=str)

    out_fname = '%s_CO2_estimation_%.1fppm_with_%s-%s.dat' % (origin, co2_ssp,
                                                              tracer, ssp)

    with open(out_fname, 'w') as f:
        wrote_header = False

        for fn in filenames:
            if origin == 'Berlin':
                data,res,_ = gauss_1d(fn, 'Berlin', show=False, detection_tracer=tracer,
                                      detection_ssp=ssp, co2_ssp=co2_ssp,
                                      load_results=False, save_results=True)
            else:
                data, res, _ = gauss_1d(fn, origin, show=False,
                        constant_emissions=False,
                         load_results=False, save_results=True, out_filename=None,
                         load_plume=False,

                         # curve
                         degree=5,
                         dmin=2e3, dmax=np.inf,
                         force_origin=True,

                         # plume detection
                         detection_tracer=tracer, detection_ssp=ssp,
                         co2_ssp=co2_ssp,
                         ns=1, q=0.99, min_pixels=10,

                         # background
                         bg_area=10, # 10x10 neighborhood

                         # wind speed
                         wind_use_mean=False, #
                         wind_hmin=120.0,      # Berlin: 0.0 m, Janschwalde: > 120 m
                         wind_hmax=1000.0      #       500.0 m,             < 3000 m
                )

            if res is None:
                continue

            if wrote_header == False:
                f.write('\t'.join(res.keys()))
                f.write('\n')
                wrote_header = True

            f.write('\t'.join([str(res[k]) for k in res.keys()]))
            f.write('\n')
            print(data.time, data.orbit, data.lon_eq, '%.1f, %.1f' % (
                  res['est. CO2 emissions (Mt/yr)'],
                  res['true CO2 emissions 11 UTC (Mt/yr)']
            ))


def load_1d_results(filename):
    data = xarray.open_dataset(filename + '_data.nc')
    data.attrs['time'] = datetime.strptime(data.time, '%Y-%m-%d %H:%M:%S')
    curve = Poly2D.from_file(filename + '_curve.npz')

    results = OrderedDict()
    with open(filename + '_results.dat', 'r') as f:
        for line in f:
            key, value = line.split('\t')
            if key in ['time', 'plume_detection']:
                results[key] = value.strip()
            elif key in ['orbit', 'lon_eq']:
                results[key] = int(value)
            else:
                results[key] = float(value)

    return data, results, curve


def results2timeseries(path):
    """
    Combine results from single mass-balance estimates to pandas.DataFrame.
    """
    pattern = os.path.join(path, '*_results.dat')
    fnames = sorted(glob.glob(pattern))

    for i,fn in enumerate(fnames):
        with open(fn) as lines:
            c = dict(line.strip().split('\t') for line in lines)

        index = datetime.strptime(c['time'], '%Y-%m-%d %H:%M:%S')
        del c['time']

        for k,v in c.items():
            try:
                c[k] = float(v)
            except ValueError:
                pass

        if i == 0:
            frame = pandas.DataFrame(c, columns=c.keys(), index=[index])
        else:
            frame = frame.append(pandas.DataFrame(c, columns=c.keys(), index=[index]))

    return frame


def plume_signal_and_gaps(data, tracer='CO2', use_true_background=False,
                          fill_gaps=False):
    """
    compute plume signal and fill data gaps
    """

    if use_true_background:
        name = '%s_minus_true_background' % tracer
        signal = data[tracer] - data['%s_true_background' % tracer]
    else:
        name = '%s_minus_estimated_background' % tracer
        signal = data[tracer] - data['%s_estimated_background' % tracer]

    # fill data gaps
    if fill_gaps:
        print('Fill gaps!')
        gaps = data.plume_area & np.isnan(signal)
        signal = fill_gaps(signal, gaps, data.plume_area)

    data[name] = DataArray(signal, dims=data[tracer].dims,
                           attrs=data[tracer].attrs)
    return data



def draw_sketch(data, curve):

    x = data.x.values[data.detected_plume]
    y = data.y.values[data.detected_plume]

    fig = plt.figure()
    ax = fig.add_subplot(111)#, aspect=1)

    ax.plot(x,y, 'k+', label='detected pixels')
    ax.plot(curve.x_o, curve.x_o, 'ko', label='source')
    ax.plot(curve.x0, curve.y0, 'kD', label='origin')
    
    ax.pcolormesh(data.x, data.y, data.CO2_noisefree, vmin=405, vmax=406)


    tmax = curve.get_parameter(x, y).max() + 20e3
    t = np.linspace(0, tmax)

    ax.plot(*curve(t=t), ls='-', color='k')


    #px, py = -2e4, 0.0
    #qx, qy = -1e4, 0.0 # TODO
    #ax.plot(px, py, 'ro')
    #ax.plot([px, qx], [py, qy], 'b-') # yp
    #ax.plot([curve.x0, px], [curve.y0, py], 'b--') # tp
    
    plt.legend()





def show_setup(data, curve, results=None, tracer=None, source='Berlin', names=None,
               add_wind=True, vlim=(-2,+2), simple_legend=True, add_multiple_sources=True,
               figwidth=6.97, cmap=None, marker='+', markersize=3):

    fig = dplume.vis.visualize(data, do_zoom=True, show_clouds=True,
                               is_sentinel_5=False, data7=None, tracer=tracer,
                               show_true_plumes=False, source=source, names=names,
                               add_wind=add_wind, vlim=vlim, simple_legend=simple_legend,
                               add_multiple_sources=add_multiple_sources, figwidth=figwidth,
                               cmap=cmap, marker=marker, markersize=markersize,
                               domain=smartcarb.cfg.DOMAINS['smartcarb']
                              )


    ax = fig.get_axes()[0]

    if curve is not None:

        #x = data.x.values[data.detected_plume]
        #y = data.y.values[data.detected_plume]
        #tmax = curve.get_parameter(x, y).max() + 20e3
        #t = np.linspace(0, tmax)
        
        along = get_distances(data, delta=10e3, min_distance=-30e3)
        tmax = curve.arc2parameter(along[-1])
        t = np.linspace(curve.t_o, tmax)            

        map.add_curve(ax, *curve(t=t),
                      color='black', label='Center line')

        #map.add_curve(ax, curve.x0, curve.y0, marker='D',
        #              ls='', color='yellow', mec='k', ms=5,
        #              label='Origin ($x_o$, $y_o$)')
        
        ya, yb = mass.get_plume_width(data, dy=5e3)
        ya = np.full_like(along, ya)
        yb = np.full_like(along, yb)
            
        mass.add_area(ax, data, curve, xa=along-5e3, xb=along+5e3, ya=ya, yb=yb)

    return fig


def detect_plume(filename, origin, tracer='CO2', show=False, constant_emissions=False,
                 load_plume=False,

                 # plume detection
                 detection_tracer='NO2', detection_ssp='high', co2_ssp=0.7,
                 background='bg-obs_var-fix',
                 ns=37, q=0.99, min_pixels=1, use_s5=False,
                 seed='orbit'):
    """
    CO2 emission estimates with Gauss-1D plume (cross section in plume)
    """

    # detect plume
    if load_plume:
        raise NotImplementedError
    else:
        cloud_thr = 0.3 if detection_tracer == 'NO2' else 0.01
        data = dplume.plume_detection_algorithm(
            filename, detection_tracer, detection_ssp, origin=origin,
            ns=ns, q=q, use_s5=use_s5, cloud_thr=cloud_thr,
            show=show, use_constant_emissions=constant_emissions,
            co2_ssp=co2_ssp,
            background=background,
            seed=seed               # fixed seed depending on orbit and lon_eq
        )

        if 'detected_plume' not in data or np.sum(data.detected_plume) < min_pixels:
            return data

    return data



def prepare_estimate_emissions(filename_or_data, sources, detection_tracer=None,
                               detection_ssp='high', background='bg-obs_var-fix',
                               co2_ssp=0.7, show=False, degree=2,
                               force_origin=False, ns=37, add_3d_wind=False,
                               plume_area=50e3):
    """
    """

    if isinstance(filename_or_data, str):
        if detection_tracer == 'NO2':
            print('Detect plume with %s (noise: %s)' % (detection_tracer,
                                                        detection_ssp))
        else:
            print('Detect plume with %s (noise: %.1f)' % (detection_tracer,
                                                          detection_ssp))
        data = detect_plume(filename, sources=sources,
                            detection_tracer=detection_tracer,
                            detection_ssp=detection_ssp, co2_ssp=co2_ssp, ns=ns,
                            background=background)
    else:
        data = filename_or_data
        
    if sources is None:
        sources = data.source.values

    if 'detected_plume' not in data:
        return None, None

    # x-y coords do not subtract the source location anymore!
    print('Compute EPSG %s coords' % cfg.EPSG)
    data = compute_xy_coords(data)

    print('Compute plume area (detection + %.0f km)' % (plume_area/1e3))
    for source in sources:
        data = compute_plume_area(data, distance=plume_area, source=source, add_to_dataset=True)
        
    for tracer in [name for name in ['CO2', 'NO2'] if name in data]:

        # add mask for valid pixels
        print('Add %s isfinite mask' % tracer)
        data['%s_isfinite' % tracer] = np.isfinite(data[tracer])
        
        # estimate XCO2 background using normalized convolution
        print('Estimate %s background' % tracer)
        data = emiss.background.estimate(data, tracer, show=show)

        # compute CO2 signal and fill gaps
        print('Estimate plume signal and fill gaps')
        if '%s_true_background' % tracer in data:
            data = plume_signal_and_gaps(data, tracer=tracer, use_true_background=True)
            
        data = plume_signal_and_gaps(data, tracer=tracer, use_true_background=False)
    
        # field with estimated background
        if '%s_noisefree_minus_estimated_background' % tracer in data:
            data['%s_noisefree_minus_estimated_background' % tracer] = (
                data['%s_noisefree' % tracer] - data['%s_estimated_background' %
                                                     tracer]
            )

        # convert ppm to kg/m2
        print('Convert to kg m-2')
        fields = [v % tracer for v in ['%s', '%s_noise', '%s_noisefree',
                                       '%s_true_signal',
                                       '%s_minus_estimated_background',
                                       '%s_noisefree_minus_estimated_background',
                                       '%s_minus_true_background',
                                       '%s_estimated_background',
                                       '%s_true_background'] if v % tracer in data]

        if tracer == 'NO2':
            data = convert_molec_to_mass(data, fields)
        else:
            data = convert_ppm_to_mass(data, fields)

    curves = {}

    print('Compute curves and plume coords')
    if 'xp' in data:
        data['xp'][:] = np.nan
        data['yp'][:] = np.nan
        
    if 'xcp' in data: 
        data['xcp'][:] = np.nan
        data['ycp'][:] = np.nan
    
    for source in sources:

        # select source
        source_data = data.sel(source=source)
        
        # compute curve
        if source_data['detected_plume'].sum() == 0 or emiss.misc.has_multiple_sources(data, source):
            curves[source] = None

        else:
            if 'local_NO2_mean' in source_data:
                weights = source_data['local_NO2_mean'].values / 1e15
                weights[source_data.plume_area & ~source_data.detected_plume] = 0.2
                weights[weights < 0.2] = 0.2
            else:
                weights = source_data['local_CO2_mean'].values
                weights[source_data.plume_area & ~source_data.detected_plume] = 0.05
                weights[weights < 0.05] = 0.05

            # fit curve to detected plume plus surroundings
            if source in ['Berlin']:
                source_distance = 50e3
            else:
                source_distance = 20e3

            curves[source] = create_center_line(source_data, source=source, degree=degree,
                                                source_distance=source_distance,
                                                force_source=force_origin)
           
            print(source)
            data = compute_plume_coords(data=data, curve=curves[source],
                                        source=source, do_corners=False)
            
            if add_3d_wind:
                for tracer in ['CO2', 'NO2']:
                    # add wind
                    print('Add wind vector')
                    data = wind.add_wind(data, tracer=tracer, curve=curve)


    return data, curves



def read_true_emissions(date, tracer, origin='Berlin', tmin=6, tmax=12):
    """
    Read true emissions for `tracer` for given date between [`tmin`,`tmax`].
    Units are Mt/yr for CO2 and kt/yr for NO2.
    """
    em = pandas.read_csv('/output/SMARTCARB/output/timeseries/ts_%s_emissions.dat' % tracer,
                         sep='\t', index_col=0, parse_dates=True)['%s (Mt/yr)'
                                                                 % origin]

    em = em[[pandas.Timestamp(date.strftime('%Y%m%d')) + timedelta(hours=h)
             for h in range(tmin,tmax+1)]]

    # NO2: Mt/yr -> kt/yr
    if tracer == 'NO2':
        em = 1e3 * em

    return em


def get_distances(data, delta, min_distance=-30e3):
    dmax = np.nanmax(data.xp.values[data.detected_plume])
    return np.arange(min_distance, dmax+delta, delta)



def rot2geo_winds(lon, lat, u, v, pollon, pollat):
    """
    based on amrs implementation (TODO: update amrs and use shared code)
    """
    pollon = np.deg2rad(pollon)
    pollat = np.deg2rad(pollat)
    lon    = np.deg2rad(lon)
    lat    = np.deg2rad(lat)
    
    delta = np.arctan( (np.cos(pollat)*np.sin(pollon-lon)) / ( (np.cos(lat)*np.sin(pollat)) -
                       (np.sin(lat)*np.cos(pollat)*np.cos(pollon-lon)) ) )

    ug = u * np.cos(delta) + v * np.sin(delta)
    vg = -u * np.sin(delta) + v * np.cos(delta)
    
    return ug, vg



def compute_angle_between_curve_and_wind(source, data, curve, hmin=0.0, hmax=1e3):
    """
    Compute the angle between wind vector and curve tangent, which can be
    used as a warning flag for large misfits.
    
    Parameter: 
    - source: name of point source
    - curves: dict with curves
    """
    # convert local reference system to lon, lat
    u, v = curve.compute_tangent(curve.t_o)
    u, v = smartcarb.misc.wgs2epsg(np.array([curve.x_o, curve.x_o + u]),
                                   np.array([curve.y_o, curve.y_o + v]),
                                   inverse=True)
    u = np.diff(u)
    v = np.diff(v)

    curve_angle = float( np.rad2deg(np.arctan2(u,v)) )

    # compute wind vector (convert from rotpole to wgs84)
    u, v = emiss.gauss.get_wind_vector(pandas.Timestamp(data.time.values),
                                       hmin=hmin, hmax=hmax, origin=source,
                                       pollon=smartcarb.cfg.pollon,
                                       pollat=smartcarb.cfg.pollat)

    wind_angle = float( np.rad2deg(np.arctan2(u,v)) )
    
    return smallest_angle(wind_angle, curve_angle) #float(np.abs(curve_angle - wind_angle) % 360)


def smallest_angle(x,y):
    return min(abs(x - y), 360 - abs(x - y))
    


def get_wind_vector(time, origin, hmin=0.0, hmax=1e3, pollon=None, pollat=None):
    """
    Get wind vector in rotated pole coords
    """
    lon, lat, _ = misc.get_source_origin(origin, data=None)
    
    w = get_wind_speed(time, hmin=hmin, hmax=hmax, origin=origin)
    speed = w['mean wind speed (m/s)']
    direction = w['mean wind direction (deg)']

    u = -1.0 * speed * np.sin(np.deg2rad(direction))
    v = -1.0 * speed * np.cos(np.deg2rad(direction))

    u = u / np.sqrt(u**2 + v**2)
    v = v / np.sqrt(u**2 + v**2)
    
    if pollon is not None and pollat is not None:
        u,v = rot2geo_winds(lon, lat, u, v, pollon, pollat)
    
    return u, v


def get_wind_speed(time, hmin=0.0, hmax=500.0, origin='Berlin', show=False):
    layers,uwind,vwind = smartcarb.io.read_wind_profile(origin, time)
    wind_info = estimate_wind_speed(layers, uwind, vwind, curve=None, show=show,
                                    use_mean=True, hmin=hmin,
                                    hmax=hmax)
    return wind_info




def estimate_emissions(data, curve, tracer='CO2', origin='Berlin', delta=10e3, show=False,
                       method='sub-areas', dmin_fit=10e3, dmax_fit=np.inf,
                       only_observations=False, wind_interval=(0.0,500.0)):
    """
    Estimate CO2/NOX emissions from detected plume.

    delta - along-track box sizes (default 10 km)
    """

    # TODO:
    # - save detected plume plus setup for flux inversion (20 realisations)
    # - test with more data
    # - test for one satellite
    # - modify for Sentinel-5
    # - apply to all data (plume detection for only NO2 data)

    if np.all(curve.c == 0):
        print('Curve fit failed.')
        return None

    if data['detected_plume'].sum() == 0:
        print('Number of detected pixels is zero.')
        return None

    results = xarray.Dataset(attrs=data.attrs)

    time = pandas.Timestamp(data.attrs['time'])

    # mean wind between 0 - 1 km in Berlin
    wind_info = get_wind_speed(data.attrs['time'], hmin=wind_interval[0],
                               hmax=wind_interval[1], origin=origin)
    u_1km = wind_info['mean wind (m/s)']

    results.attrs['wind_speed_%d-%dm' % wind_interval] = u_1km
    results.attrs['wind_direction_%d-%dm' % wind_interval] = wind_info['est. wind direction (deg)']
    results.attrs['wind_info'] = (
        'wind speed and direction from %d to %d m in m/s and degrees' % wind_interval
    )

    # along plume distances (in meters)
    distances = get_distances(data, delta)
    xvalues = (distances[1:] - delta / 2)


    name = 'along_plume_distance'
    attrs = {
        'long_name': 'along plume distance',
        'units': 'm',
    }
    results[name] = xarray.DataArray(xvalues, dims=('along'),
                                     attrs=attrs)

    # true emissions (10-11 UTC)
    em = read_true_emissions(time, tracer, tmin=6, tmax=12, origin=origin)
    q_true = np.mean(em[(em.index.hour == 10) | (em.index.hour == 11)])


    name = '%s_emissions' % tracer
    attrs = {
        'long_name': '%s emissions' % tracer,
        'units': 'kt yr-1' if tracer == 'NO2' else 'Mt yr-1',
    }
    results[name] = xarray.DataArray(em, dims='time', attrs=attrs)

    # plot general figure
    if show:
        ax, ax2, lines = vis.init_mass_balance_plot(tracer, distances, delta,
                                                    emissions=em,
                                                    origin=origin,
                                                    q_true=q_true, u_1km=u_1km,
                                                    wind_interval=wind_interval)

    if only_observations:
        iterator = zip(
            ['%s_minus_estimated_background_mass'],
            [True]
        )
    else:
        iterator = zip(
            [
                '%s_true_signal_mass',                          # mass true
                '%s_true_signal_mass',                          # only mass-balance errors
                '%s_true_signal_mass',                          # wind error
                '%s_minus_true_background_mass',                # instrument noise
                '%s_noisefree_minus_estimated_background_mass', # background error
                '%s_minus_estimated_background_mass'            # all uncertainties
            ],
            [
                False, # use mean wind speed <1km
                False,
                True,
                False,
                False,
                True
            ]
        )


    for i, (varname, use_est_wind) in enumerate(iterator):

        use_true_wind = (i == 0) and not only_observations

        flux, flux_std, ueff = estimate_fluxes(data, curve, varname % tracer,
                                               distances, u_1km=u_1km, method=method,
                                               use_est_wind=use_est_wind,
                                               use_true_wind=use_true_wind)

        if use_est_wind: # use 0-1 km mean wind
            name = 'flux_from_%s_using_mean_wind' % varname % tracer
        else:
            if i == 0:
                name = 'flux_from_model_tracer_using_true_wind'
            else:
                name = 'flux_from_%s_using_effective_wind' % varname % tracer

        # fit 
        q, q_std, fit, _, _ = fit_emissions(tracer, xvalues, flux,
                                            flux_std=None,
                                            dmin_fit=dmin_fit, dmax_fit=dmax_fit)

        name = name.replace('_mass', '')
        results[name] = xarray.DataArray(flux, dims='along')


        attrs = {
            'units': 'Mt CO2 yr-1' if tracer == 'CO2' else 'kt NO2 yr-1',
            'emissions': q,
            'emissions_std': q_std,
            'true_emissions_10-11UTC': q_true
        }
        results[name+'_fit'] = xarray.DataArray(fit, dims='along', attrs=attrs)

        results.attrs['wind_speed_eff'] = np.mean(ueff)
        results.attrs['wind_eff_info'] = 'effective wind speed in plume (m s-1)'

        if show != False:

            if show is True:
                add_result = True
            elif show == 'some' and i > 0:
                add_result = True
            elif show == 'combined' and i == 5:
                add_result = True
            elif show == 'model+total' and i in [1, 5]:
                add_result = True
            else:
                add_result = False

            lines = vis.add_mass_balance_plot(i, ax, ax2, tracer, distances, delta, flux, flux_std,
                                              ueff, lines=lines,
                                              add_result=add_result,
                                              add_wind=(i==0), dq=(q-q_true),
                                              fit=fit)

    if show:
        vis.finalize_mass_balance_plot(ax, ax2, tracer, lines)


    return results


def estimate_fluxes(data, curve, varname, distances, method, u_1km=None, use_est_wind=True,
                    use_true_wind=False):
    """
    Estimate fluxes in along-plume direction.
    """
    xvalues = (distances[1:] - np.diff(distances) / 2)


    # mass, mass_std and flux from true wind
    valids = data['%s_isfinite' % varname[:3]].values
    m, m_std, m_true, ueff, u_true = mass.integrate_along_line(data, varname,
                                                               distances, curve=curve,
                                                               show=False,
                                                               method=method)
    ueff[:] = np.nanmean(ueff[xvalues >= 10e3])

    if use_true_wind:
        m = m_true
        ueff = u_true

    # convert NO2 to NOX (Beirle et al. 2011)
    if varname.startswith('NO2'):
        m *= 1.32
        m_std *= 1.32
        m_true *= 1.32

    # uncertainty wind speed: 2 m/s

    if use_est_wind: # use 0-1 km mean wind
        flux = m * u_1km
        flux_std = np.sqrt(m_std**2 * u_1km**2 + 2.0**2 * m**2)
    else:
        flux = m * ueff
        flux_std = np.sqrt(m_std**2 * ueff**2 + 2.0**2 * m**2)


    flux_std[flux_std < 0.3] = 0.3


    return flux, flux_std, ueff






def gaussian(x, sigma):
    g = 1 / (np.sqrt(2.0 * np.pi) * sigma) * np.exp(-x**2 / (2.0 * sigma**2))

    return g


def decay_function(x, x_source, x0=None):
    e = np.zeros_like(x)
    downstream = x > x_source

    if x0 is None:
        e[downstream] = 1.0
    else:
        e[downstream] = np.exp(-(x[downstream] - x_source) / x0)

    return e


def plume_model(x, Q, sigma, x0=None, x_source=0.0, B=0.0, dx=1e3):
    """
    x : along-plume distance (in meters)
    sigma: width of gaussian
    x_source : plume source
    x0: decay distance

    B: background
    """
    # high-resolution x-distance
    xhigh = np.arange(x[0]-50e3, x[-1]+200e3, dx)

    # decay function
    e = decay_function(xhigh, x_source, x0)

    # gaussian
    xgauss = np.arange(-50e3, +50e3+dx/2, dx)
    g = gaussian(xgauss, sigma)

    # convolution
    f = ndi.convolve(e, g, mode='nearest')

    # scaling with source strength assumed
    M = Q * f / f.max() + B

    # interpolate
    M = np.interp(x, xhigh, M)

    return M


def point_plume_model(x, Q, x0):
    e = Q * decay_function(x, 0.0, x0)
    return e


def fit_emissions(tracer, xvalues, flux, flux_std=None,
                  dmin_fit=10e3, dmax_fit=np.inf, is_point_source=True,
                  absolute_sigma=True):
    """
    Estimate NO2 emissions from fluxes by fitting an exponential decay
    function.
    """
    # check valid values
    valids = np.isfinite(flux) & (xvalues < dmax_fit)
    valids[xvalues < dmin_fit] = False

    if flux_std is None or np.all(np.isnan(flux_std)) or np.all(flux_std == 0.0):
        sigma = None
    else:
        sigma = flux_std[valids]

    # starting vector
    if tracer == 'NO2':
        
        if is_point_source:
            p0 = np.array([
                10.0,    # source strength in kg/s (~10 for NOx and 1000 for CO2)
                100e3,   # decay distance: x0 = wind_speed * decay_time
            ])
            model = point_plume_model
            bounds= [0.0, 0.0], [np.inf, 432000.0] #24 hours
            
            
            
        else:
            p0 = np.array([
                10.0,    # source strength in kg/s (~10 for NOx and 1000 for CO2)
                10e3,    # width of city gaussian
                100e3,   # decay distance: x0 = win_speed * decay_time
            ])
            model = plume_model
            bounds= [0.0, 0.0, 0.0], [np.inf, np.inf, np.inf]

    else:
        p0 = np.array([
            1000.0, # source strength in kg/s (~10 for NOx and 1000 for CO2)
        ])
        model = lambda x, q: np.full(x.shape, q)
        bounds= [0.0], [np.inf]


    if np.sum(valids) < {'NO2': 1, 'CO2': 1}[tracer]:
        print('Too few valid line densities (%d) for fitting curve.' %
              sum(valids))

        return np.nan, np.nan, np.full(xvalues.shape, np.nan), np.full_like(p0, np.nan), np.full_like(p0, np.nan)
        
    
    if tracer == 'CO2' and np.sum(valids) == 1:
        x = np.array([flux[valids][0]])
        x_std = np.array([flux_std[valids][0]])
        
    elif tracer == 'NO2' and np.sum(valids) < 3 and np.any(xvalues[valids] < 10e3):
        x = np.array([np.mean(flux[valids]), np.nan])
        x_std = np.array([np.mean(flux_std[valids]), np.nan])
        
    else:
        try:
            x, cov_x = scipy.optimize.curve_fit(model, xvalues[valids],
                                                flux[valids], p0=p0,
                                                bounds=bounds, 
                                                sigma=sigma, absolute_sigma=absolute_sigma)

        except RuntimeError as e:
            print('Fit fluxes:', e)
            return np.nan, np.nan, np.full(xvalues.shape, np.nan), np.full_like(p0, np.nan), np.full_like(p0, np.nan)
    
        except scipy.optimize.OptimizeWarning as e:
            print('Fit fluxes:', e)
            return np.nan, np.nan, np.full(xvalues.shape, np.nan), np.full_like(p0, np.nan), np.full_like(p0, np.nan)

            


        # estimate uncertainty assuming a good fit
        x_std = np.sqrt(np.diag(cov_x))

    if tracer == 'NO2':
        if is_point_source:
            if np.isnan(x[1]): # no decay time
                fit = np.full(xvalues.shape, x[0])
            else:
                fit = point_plume_model(xvalues, *x)
        else:
            fit = plume_model(xvalues, *x)
    else:
        fit = np.full(xvalues.shape, np.nan)
        fit[xvalues >= dmin_fit] = x[0]

    fit = kgs_to_Mtyr(fit)

    # in kt yr-1 or Mt yr-1
    if tracer == 'NO2':
        q = 1e3 * kgs_to_Mtyr(x[0])
        q_std = 1e3 * kgs_to_Mtyr(x_std[0])
        fit = 1e3 * fit
        
    else:
        q = kgs_to_Mtyr(x[0])
        q_std = kgs_to_Mtyr(x_std[0])
        

    return q, q_std, fit, x, x_std



def save_1d_results(filename, data, results, curve):

    data.attrs['time'] = str(data.attrs['time'])
    try:
        data.to_netcdf(filename + '_data.nc')
    except PermissionError:
        data.to_netcdf(filename + '_data.nc.tmp')
        data.close()
        shutil.move(filename + '_data.nc.tmp', filename + '_data.nc')

    with open(filename + '_results.dat', 'w') as f:
        for key, value in results.items():
            f.write('%s\t%s\n' % (key, str(value)))

    curve.to_file(filename + '_curve.npz')



def bg_std(filenames):
    results = []

    for filename in filenames:
        data = xarray.open_dataset(filename + '_data.nc')
        data.attrs['time'] = datetime.strptime(data.time, '%Y-%m-%d %H:%M:%S')
        print(data.attrs['time'])

        v = np.array(data.CO2)
        v[data.is_hit] = np.nan
        v[~data.plume_area] = np.nan

        r = (data.attrs['time'], data.orbit, data.lon_eq, np.nanstd(v))
        results.append(r)

    return results


def all_gauss_2d(tracer, ssp, co2_ssp, origin='Janschwalde', degree=5,
                 outpath='.', bg_area=10):

    if tracer == 'CO2':
        filename = (
            '/project/kug/smartcarb/results/Janschwalde'
            '/plume_detection/CO2_%.1f_q0.990_ns1.dat' % ssp
        )
    else:
        filename = (
            '/project/kug/smartcarb/results/Janschwalde'
            '/plume_detection/NO2_high_q0.990_ns1.dat'
        )
    d = pandas.read_csv(filename, sep='\t',
                        parse_dates=True, index_col=0)
    filenames = smartcarb.io.table2filename(d)[d.CO2_pixels >= 10]

    out_fname = '%s_CO2_estimation_%.1fppm_with_%s-%s_d%s.dat' % (
        origin, co2_ssp, tracer, ssp, degree)

    with open(out_fname, 'w') as f:
        wrote_header = False

        for i,fn in enumerate(filenames):
            print(i,os.path.basename(fn))
            data,res,_,_ = gauss_2d(fn, origin, show=True, detection_tracer=tracer,
                                    degree=degree,
                                    detection_ssp=ssp, co2_ssp=co2_ssp,
                                    load_results=False, save_results=True,
                                    use_mean=True, outpath=outpath,
                                    bg_area=bg_area)

            plt.close('all')

            if data is None:
                continue

            if wrote_header == False:
                f.write('\t'.join(res.keys()))
                f.write('\n')
                wrote_header = True

            f.write('\t'.join([str(res[k]) for k in res.keys()]))
            f.write('\n')
            print(i,data.time, data.orbit, data.lon_eq,
                  '%.1f' % res['est. CO2 emissions (Mt/yr)'],
                  '%.1f' % res['true CO2 emissions 11 UTC (Mt/yr)'],
            )



def gauss_2d(filename, origin, degree=5, offset=0, show=False,
             detection_tracer='NO2', detection_ssp='high', ns=1, co2_ssp=0.7,
             constant_emissions=False, pre_est_bg=True, load_results=False,
             save_results=False, use_mean=True, outpath='.', bg_area=10):

    # out filename prefix
    out_filename = os.path.basename(filename)
    out_filename = os.path.splitext(out_filename)[0]

    # detect plume
    cloud_thr = 0.3 if detection_tracer == 'NO2' else 0.01
    data = dplume.plume_detection_algorithm(
        filename, detection_tracer, detection_ssp, origin=origin,
        ns=ns, cloud_thr=cloud_thr, show=False,
        use_constant_emissions=constant_emissions, co2_ssp=co2_ssp
    )

    if np.sum(data.detected_plume) < 10:
        return None, None, None, None

    results = OrderedDict()
    results['time'] = data.attrs['time']
    results['orbit'] = data.orbit
    results['lon_eq'] = data.lon_eq
    results['plume size'] = int(np.sum(data.detected_plume))
    results['plume_detection'] = data.tracer # tracer used for plume detection

    # setup flux inversion
    data, curve = setup_flux_inversion(data, origin=origin, degree=degree,
                                       area_size=bg_area)

    if pre_est_bg:
        # estimate XCO2 background using normalized convolution
        data = estimate_background(data, 'CO2', show=False)
        CO2_signal = np.array(data.CO2 - data.CO2_background)
    else:
        CO2_signal = np.array(data.CO2)

    CO2_signal[~data.plume_area] = np.nan
    data['CO2_signal'] = DataArray(CO2_signal, dims=data.CO2.dims,
                                   attrs=data.CO2.attrs)

    # wind speed
    layers,uwind,vwind = smartcarb.io.read_wind_profile(origin, data.attrs['time'])
    params = estimate_wind_speed(layers, uwind, vwind, curve, show=True,
                                 use_mean=False, hmin=100.0, hmax=3e3)

    #if save_results:
    #    plt.savefig(out_filename + '_wind.png')

    for key in sorted(params):
        results[key] = params[key]

    results['plume length (km)'] = 1e-3 * np.nanmax(data.xp)
    results['plume age (hours)'] = (
        np.nanmax(data.xp) / results['est. wind (m/s)'] / 3600
    )

    # convert ppm to kg/m2
    data = convert_ppm_to_mass(data, ['CO2', 'CO2_noise', 'CO2_signal'])


    # setup inversion with 2D Gauss
    area = data.plume_area & np.isfinite(data.CO2_signal_mass)
    if use_mean:
        xp = np.array(data.xcp)[area]
        yp = np.array(data.ycp)[area]
        zp = np.array(data.CO2_signal_mass)[area]
    else:
        xp = np.array(data.xp)[area]
        yp = np.array(data.yp)[area]
        zp = np.array(data.CO2_signal_mass)[area]

    if pre_est_bg:
        p = [1500, 3000]
    else:
        p = [1500, 3000, zp.mean()]

    ssp = data.CO2_signal_mass.attrs['noise']
    res = flux_inversion(p, xp, yp, zp, params['est. wind (m/s)'], ssp,
                         use_mean)

    # TODO: uncertainty of wind speed
    unc = scipy.linalg.inv(np.dot(res.jac.T, res.jac))
    unc = np.sqrt(unc.diagonal())

    results['est. CO2 emissions (Mt/yr)'] = kgs_to_Mtyr(res.x[0])
    results['est. CO2 emissions uncertainty (Mt/yr)'] = kgs_to_Mtyr(unc[0])
    results['eddy diffusion coefficient (m2/s)'] = res.x[1]
    results['eddy diffusion coefficient uncertainty (m2/s)'] = unc[1]

    if not pre_est_bg:
        results['est. CO2 background (kg/m2)'] = res.x[2]
        results['est. CO2 background uncertainty (kg/m2)'] = unc[2]

    true_emissions = smartcarb.io.read_emissions(origin, data.attrs['time'],
                                       use_constant=constant_emissions)[0]
    results['true CO2 emissions 11 UTC (Mt/yr)'] = true_emissions
    results['message'] = res['message']

    if show:
        area = data.plume_area
        xp = np.array(data.xp)
        yp = np.array(data.yp)
        c = gauss2d(res.x, xp, yp, params['est. wind (m/s)'])
        c = ppm_to_mass(c, data.psurf, 'CO2', inverse=True)

        v = data.CO2 - np.mean(data.CO2_background)
        z = np.array(data.plume_area)
        plot_plume(data, curve, v, origin=origin, zoom_area=z,
                   add_cross_sections=False, contour=c, vmin=-2, vmax=4)
        if save_results:
            plt.savefig(os.path.join(outpath, out_filename + '_plume.png'))

    if save_results:

        if detection_tracer == 'NO2':
            out_filename = '%s_%s-%s_and_%.1f' % (out_filename, detection_tracer,
                                         detection_ssp, co2_ssp)
        else:
            out_filename = '%s_%s-%.1f' % (out_filename, detection_tracer,
                                           detection_ssp)

        save_1d_results(os.path.join(outpath, out_filename), data, results,
                        curve)

    return data, results, curve, res


def get_plume_size_table():
    """
    Get table with NO2 plume size and columns stating if emission could be
    estimated by mass balance approach.
    """
    ts = pandas.read_csv(
        "/output/SMARTCARB/output/timeseries/NO2_Berlin_plume_size_mass_balance.dat",
        comment='#', sep='\t', parse_dates=[0], dayfirst=True
    )
    return ts



def get_filenames(which='all'):

    ts = get_plume_size_table()

    if which == 'NO2':
        ts = ts[ts.mass_balance_suitable]

    if which == 'CO2':
        ts = ts[ts.CO2_mass_balance_suitable]

    filenames = []

    for number in range(ts.shape[0]):
        date = ts.iloc[number].date
        orbit = ts.iloc[number].orbit
        lon_eq = ts.iloc[number].lon_eq

        filename = smartcarb.io.create_filename(date, input_files='level2',
                                                instrument='Sentinel_7_CO2',
                                                orbit=orbit, lon_eq=lon_eq)

        filenames.append(filename)

    return filenames




def prepare(filename, overwrite=False, save=True, origin='Berlin',
            co2_ssp='medium', no2_ssp='high', background='bg-obs_var-fix',
            ns=37, force_origin=False):

    output_path = emiss.io.get_output_path(co2_ssp, no2_ssp, origin)

    # detection tracer and its single sounding precision
    co2_ssp = {'low': 0.5, 'medium': 0.7, 'high': 1.0}[co2_ssp]
    detection_tracer = {'high': 'NO2', 'none': 'CO2'}[no2_ssp]
    detection_ssp = {'CO2': co2_ssp, 'NO2': no2_ssp}[detection_tracer]

    basename = os.path.join(output_path, os.path.basename(filename))

    data_filename = basename.replace('.nc', '_data.nc')
    curve_filename = basename.replace('.nc', '_curve.nc')

    if not overwrite and os.path.exists(data_filename):
        print('Skip:', os.path.basename(filename))
        return None, None

    img_basename = os.path.join(output_path, 'images',
                                os.path.basename(filename))
    co2img_filename = img_basename.replace('.nc', '_co2image.png')
    no2img_filename = img_basename.replace('.nc', '_no2image.png')

    # detect plume and prepare estimate emissions
    data, curve = prepare_estimate_emissions(filename, origin=origin,
                                             detection_tracer=detection_tracer,
                                             detection_ssp=detection_ssp,
                                             background=background, ns=ns,
                                             co2_ssp=co2_ssp,
                                             force_origin=force_origin,
                                             show=False)

    if data is None:
        return None, None


    # if detection or cloud-free CO2/NO2 plume make image and save
    if not ((np.any(data['detected_plume']) or
            np.any(data['CO2_isfinite'] & data['CO2_true_plume']) or
            np.any(data['NO2_isfinite'] & data['NO2_true_plume']))):

        return None, None


    # data to file
    if save and curve is not None:
        curve.to_file(curve_filename)

    show_setup(data, curve, None, tracer='CO2')

    if save:
        plt.savefig(co2img_filename)
    plt.close()

    if detection_tracer == 'NO2':
        show_setup(data, curve, None, tracer='NO2')

        if save:
            plt.savefig(no2img_filename)
        plt.close()

    data.attrs['time'] = str(data.attrs['time'])

    if save:
        data.to_netcdf(data_filename)

    return data, curve



def load_prepared(filename, co2_ssp='medium', no2_ssp='high', origin='Berlin'):
    """
    Load prepared dataset and curve with plume detection.
    """
    sat, orbit, time, lon_eq = smartcarb.io.process_filename(filename)

    output_path = emiss.io.get_output_path(co2_ssp, no2_ssp, origin=origin)
    basename = '%s_%s_o%04d_l%04d' % (sat, time.strftime('%Y%m%d%H'), orbit, lon_eq)
    basename = os.path.join(output_path, basename)

    data_filename = basename + '_data.nc'
    curve_filename = basename + '_curve.nc'

    data = xarray.open_dataset(data_filename)
    data.attrs['time'] = pandas.Timestamp(data.attrs['time'])
    curve = Poly2D.from_file(curve_filename)

    return data, curve



def estimate(tracer, filename, method, co2_ssp='medium', no2_ssp='high'):
    try:
        data, curve = load_prepared(filename, co2_ssp=co2_ssp, no2_ssp=no2_ssp)
    except FileNotFoundError:
        return


    if not emiss.io.is_good_overpass(data.time, co2_ssp, no2_ssp):
        return


    output_path = emiss.io.get_output_path(co2_ssp, no2_ssp, origin=origin)
    basename = os.path.join(output_path, os.path.basename(filename))
    img_basename = os.path.join(output_path, 'images',
                                os.path.basename(filename))

    # estimate emissions
    results = estimate_emissions(data, curve, tracer=tracer,
                                 show=True, delta=10e3, method=method)

    if results is None:
        return

    results.attrs['time'] = str(results.attrs['time'])

    results_filename = basename.replace('.nc', '_%s_%s-results.nc' % (method,
                                                                      tracer))
    results.to_netcdf(results_filename)

    image_filename = img_basename.replace('.nc', '_%s_%s-results.png' %
                                          (method, tracer))
    plt.savefig(image_filename)
    plt.close()


    # update plume detection image
    show_setup(data, curve, results, tracer=tracer)
    image_filename = img_basename.replace('.nc', '_%simage.png' %
                                          tracer.lower())
    plt.savefig(image_filename)
    plt.close()


def main_estimate(tracer, start=0, ncpu=1, method='sub-areas',
                  co2_ssp='medium', no2_ssp='high'):

    filenames = get_filenames(tracer)

    if ncpu == 1:
        for i, filename in enumerate(filenames):

            if i < start:
                continue

            print(i, filename)
            estimate(tracer, filename, method, co2_ssp=co2_ssp,
                     no2_ssp=no2_ssp)
    else:
        args = [(tracer, fn, method, co2_ssp, no2_ssp)
                for fn in filenames[start:]]

        with Pool(ncpu) as p:
            p.starmap(estimate, args)



def main_prepare(start=0, ncpu=1, overwrite=False, co2_ssp='medium',
                 no2_ssp='high'):
    """
    """
    #filenames = get_filenames('NO2')
    filenames = sorted(np.sum([glob.glob(
        '/output/SMARTCARB/output/level2/Sentinel_7_CO2/Sentinel_7_CO2_*_*_l%04d.nc'
        % lon_eq) for lon_eq in smartcarb.misc.lon_eq_by_nsat('6')]))

    print(len(filenames))

    if ncpu == 1:

        for number, filename in enumerate(filenames):

            if number < start:
                continue

            print(number, os.path.basename(filename))

            # TODO/FIXME: use different noise patterns
            data, curve = prepare(filename, overwrite=overwrite,
                                  co2_ssp=co2_ssp, no2_ssp=no2_ssp)

    else:
        args = [(fn, overwrite, True, co2_ssp, no2_ssp) for fn in filenames[start:]]

        with Pool(ncpu) as p:
            p.starmap(prepare, args)


    # scenarios (20x)
    # - CO2 0.5 
    # - CO2 0.7
    # - CO2 1.0
    # - CO2 0.5 + NO2 (only detection)
    # - CO2 0.7 + NO2 (only detection)
    # - CO2 1.0 + NO2 (only detection)
    # - NO2 (estimate emissions)















