"""
Code for applying mass-balance approach to point sources.
"""

from collections import defaultdict
from datetime import datetime
import glob
import os
import multiprocessing as mp
import shutil
import textwrap
import warnings

from matplotlib.dates import DateFormatter
import cartopy.feature as cfeature
import matplotlib.pyplot as plt
import netCDF4
import numpy as np
import pandas
import scipy.optimize
import xarray

import smartcarb
import dplume

from . import timeprofiles as tp
from . import gauss
from . import mass
from . import misc
from . import vis

OUTPUT_PATH = '/output/SMARTCARB/output/mass-balance/PointSources'

LON_EQ_TO_LETTER = {
    0: 'a',  401: 'b', 805: 'c', 1207: 'd', 1610: 'e', 2012: 'f'
}




def plot_plume_age_and_length(data, sources, sort_by_emissions=False, scaling=1.0):
    """
    Make boxplots and histograms for plume length and age    
    """
    sources = sorted(sources) 

    # sort by emissions
    if sort_by_emissions:
        emissions = [float(data['%s-NO2-high-scaling-1.0' % source]['true_CO2_emissions'].mean()) for source in sources]
        emissions = np.array(emissions)
        sources = np.array(sources)[emissions.argsort()[::-1]]

    ages = [data[f'{source}-NO2-high-scaling-{scaling:.1f}']['plume_age'] for source in sources]
    lengths = [data[f'{source}-NO2-high-scaling-{scaling:.1f}']['plume_length'] for source in sources]
    
    
    return ages, lengths
                    
    
    
    # figure with boxplots
    fig, axes = plt.subplots(2,1, sharex=True, figsize=(6.97, 5))

    axes[0].boxplot(lengths, labels=sources)
    axes[0].set_ylim(0,250)
    axes[0].set_ylabel('Plume length [km]')

    axes[1].boxplot(ages, labels=sources)
    axes[1].set_ylim(0,20)
    axes[1].set_ylabel('Plume age [h]')


    for tick in axes[1].get_xticklabels():
        tick.set_rotation(90)

    fig.align_ylabels(axes)
    plt.tight_layout()
    
    filename = os.path.join(OUTPUT_PATH, 'NO2-high-scaling-1.0-plume_age_and_length_boxplot.pdf')
    plt.savefig(filename)
    
    # histograms
    fig, axes = plt.subplots(2,1, figsize=(6.97/2, 5))
    axes[0].hist(np.concatenate(lengths), bins=np.arange(0,251,5))
    axes[1].hist(np.concatenate(ages), bins=np.arange(0,17.1,0.5))

    axes[0].set_ylim(0,170)
    axes[1].set_ylim(0,320)

    axes[0].set_xlabel('Plume length [km]')
    axes[1].set_xlabel('Plume age [h]')

    axes[0].set_ylabel('Number of plumes')
    axes[1].set_ylabel('Number of plumes')
    fig.align_ylabels(axes)

    plt.tight_layout()
    
    filename = os.path.join(OUTPUT_PATH, 'NO2-high-scaling-1.0-plume_age_and_length_hist.pdf')
    plt.savefig(filename)
    
    
    ages = np.concatenate(ages)
        
    print('Median age:', np.median(ages))
    print('Max age:', np.max(ages))  
    print('Age < 4 hours:',sum(ages <= 4.0) / ages.size)  
    
    



def add_ratio(data):
    """
    Add CO2:NOx ratio, but don't compute the ratio's uncertainity because number of estimates
    are often too small and individual uncertainties to large to compute a correct uncertainty
    by error propagation.
    """
    data['true_CO2:NO2_emissions'] = data['true_CO2_emissions'] / data['true_NO2_emissions']
    data['est_CO2:NO2_emissions'] = data['est_CO2_emissions'] / data['est_NO2_emissions'] # Mt/kt
    data['est_CO2:NO2_emissions_std'] = xarray.full_like(data['est_CO2_emissions_std'], np.nan)
    
    # convert Mt/kt -> Mt/Mt
    data['true_CO2:NO2_emissions'] *= 1000.0
    data['est_CO2:NO2_emissions'] *= 1000.0
    data['est_CO2:NO2_emissions_std'] *= 1000.0

    return data


def add_is_okay(data, source, path):
    """
    Check for a csv file with a column with `overpass` times and a second
    column `is_okay` with true or false that is used to mask bad estimates
    based on visual inspection. If the file does not exists, nothing will
    be done.
    """
    filename = os.path.join(OUTPUT_PATH, path, 'ts_%s_is_okay.csv' % source)
    
    try:
        is_okay = pandas.read_csv(filename, index_col=0, parse_dates=True)
    except FileNotFoundError: 
        print('Could not mask bad data: %s' % filename)
        return data
        
    okay_times = is_okay.index[is_okay['is_okay'].values]

    data['is_okay'] = xarray.DataArray(np.zeros(data.overpass.size, bool), dims='overpass')

    for i, t in enumerate(data.overpass):
        if t in okay_times:
            data['is_okay'].values[i] = True
            
    is_okay = data['is_okay'].values
    
    
    data['est_CO2_emissions'].values[~is_okay] = np.nan 
    data['est_NO2_emissions'].values[~is_okay] = np.nan     
    data['est_CO2_emissions_std'].values[~is_okay] = np.nan 
    data['est_NO2_emissions_std'].values[~is_okay] = np.nan     
    
    return data



def read_timeseries(source, nsat='6', tracer='NO2', ssp='high', scaling=1.0,
                    add_extra_error=True, without_no2=False):
    """
    Create time series of estimate CO2 and NOx emissions by reading all
    netCDF files with flux estimates for a given source.
    
    Parameters
    ----------
    source           name of source (e.g. "Boxberg")
    nsat             number of satellites or constellation id (e.g. "2c", "3a", "6")
    tracer           trace gas used for plume detection (CO2 or NO2)
    ssp              single sounding precision scenario for trace gas ("low" or "high")
    scaling          scaling applied to anthropogenic NO2 signals (default: 1.0)
    add_extra_error  add estimate of uncertainty from method and background
                     that is not included in the estimated error
    """
    dataset = xarray.Dataset()
    
    for t in ['CO2', 'NO2']:
        em_true = smartcarb.io.read_emissions(origin=source, time='raw',
                                              use_constant=False,
                                              tracer=t)
        if t == 'NO2':
            em_true *= 1e3 # Mt yr-1 -> kt yr-1
            em_true *= scaling
            units = 'kt yr-1'
        else:
            units = 'Mt yr-1'

        dataset['time'] = xarray.DataArray(em_true.index, dims='time')
        dataset['true_%s_emissions' % t] = xarray.DataArray(em_true.values, dims='time')
        
        dataset['true_%s_emissions' % t].attrs['units'] = units
        dataset['true_%s_emissions' % t].attrs['species'] = t
    
    
    ts = defaultdict(list)
    
    # get filenames with flux information
    if without_no2:
        path = '%s-%s-scaling-%.1f-without-no2' % (tracer, ssp, scaling)
    else:
        path = '%s-%s-scaling-%.1f' % (tracer, ssp, scaling)
    
    pattern = os.path.join(OUTPUT_PATH, path, source, 'Sentinel_7_CO2_*_flux.nc')
    fnames = sorted(glob.glob(pattern))
    
    lon_eqs_in_constellation = smartcarb.misc.lon_eq_by_nsat(nsat)
    
    if len(fnames) == 0:
        print('No files found!')
        return None
    
    for fn in fnames:
        with xarray.open_dataset(fn) as d:
            _, orbit, time, lon_eq = smartcarb.io.process_filename(fn)
            orbit = smartcarb.io.filename2orbit(fn)
            
            if lon_eq not in lon_eqs_in_constellation:
                #print('Skip', time, lon_eq)
                continue
                
            
            ts['overpass'].append(time)
            ts['orbit'].append(orbit)
            ts['lon_eq'].append(lon_eq)
            ts['est_CO2_emissions'].append(d.CO2_flux_fit.attrs['emissions'])
            ts['est_CO2_emissions_raw_std'].append(d.CO2_flux_fit.attrs['emissions_std'])           
            ts['est_CO2_emissions_std'].append(d.CO2_flux_fit.attrs['emissions_std'])           
            ts['est_NO2_emissions'].append(d.NO2_flux_fit.attrs['emissions'])
            ts['est_NO2_emissions_raw_std'].append(d.NO2_flux_fit.attrs['emissions_std'])              
            ts['est_NO2_emissions_std'].append(d.NO2_flux_fit.attrs['emissions_std'])              
            ts['est_decay_time'].append(d.NO2_flux_fit.attrs['decay_time'])
            ts['est_decay_time_std'].append(d.NO2_flux_fit.attrs['decay_time_std'])
            ts['angle_between_curve_and_wind'].append(d.attrs['angle_between_curve_and_wind'])
            ts['wind_speed'].append(d.NO2_flux_fit.attrs['wind_speed'])
            ts['detected_pixels_upstream'].append(d.is_plume.isel(along=0).sum())
            
            plume_age, plume_length = compute_plume_age_and_length(d)
            ts['plume_age'].append(plume_age)
            ts['plume_length'].append(plume_length)

    
    for key, values in ts.items():
        dataset[key] = xarray.DataArray(values, dims='overpass')
        
    
    dataset['plume_length'].attrs.update({
        'description': 'estimated plume length based on plume detection',
        'units': 'km'
    })
    dataset['plume_age'].attrs.update({
        'description': 'estimated plume age based on plume detection and wind speed',
        'units': 'h'
    })
        
    for species in ['CO2', 'NO2']:

        # add error estimate for mass-balance approach and background
        if add_extra_error:
            slope, intercept = {
                    'CO2': (0.32, 1.26),
                    'NO2': (0.28, 0.30)                
            }[species]
            
            
            Q = vis.extract_time_range(dataset.time, dataset[f'true_{species}_emissions'],
                                       10, 11)[1].mean()
            
            extra = slope * Q + intercept
            dataset['est_%s_emissions_std' % species] = np.sqrt(dataset['est_%s_emissions_std' % species]**2 + extra**2)
            
            print('Add extra uncertainty: %.2f' % extra)
        
        bad_values = (dataset['est_%s_emissions' % species] > 1000.0)
        bad_values |= (dataset['est_%s_emissions_std' % species] > 75.0)
        
        bad_values |= dataset['angle_between_curve_and_wind'].values > 45.0
        bad_values |= dataset['detected_pixels_upstream'].values >= 5
        
        #print('Mask %d bad estimates.' % np.sum(bad_values))
        dataset['est_%s_emissions' % species].values[bad_values] = np.nan        
        dataset['est_%s_emissions_std' % species].values[bad_values] = np.nan
        
        
    #dataset = add_is_okay(dataset, source, path=path)    
    
    # add ratio
    dataset= add_ratio(dataset)
        
        
    return dataset


def compute_statistics(data, tracer, lon_eqs_in_constellation,
                       true_annual_mean=0.0, quiet=False, use_raw=False):
    """
    Compute mean bias, scatter etc for time series of estimate emissions
    or emission ratios.
    """
    valids = np.isin(data.lon_eq, lon_eqs_in_constellation)
    true_values = data['true_%s_emissions' % tracer].sel(time=data.overpass).values[valids]
    est_values = data['est_%s_emissions' % tracer].values[valids]

    n = np.isfinite(est_values).sum()
     
    mb = np.nanmean(est_values - true_values)
    sd = np.nanstd(est_values - true_values)
    nmb = 100*np.nanmean(est_values - true_values) / true_annual_mean
    nsd = 100*np.nanstd(est_values - true_values) / true_annual_mean
    rmsd = np.sqrt(mb**2 + sd**2 / n)
    nrmsd = 100 * rmsd / true_annual_mean
  
    values = est_values - true_values
    pr = 0.5 * (scipy.stats.scoreatpercentile(values[np.isfinite(values)], 50+34) - scipy.stats.scoreatpercentile(values[np.isfinite(values)], 50-34))
    npr = 100 * pr / true_annual_mean
   
    if np.any(valids):
        if use_raw:
            median = np.nanmedian(data['est_%s_emissions_raw_std' % tracer].values[valids])
            minimum = np.nanmin(data['est_%s_emissions_raw_std' % tracer].values[valids])
            maximum = np.nanmin(data['est_%s_emissions_raw_std' % tracer].values[valids])
        else:
            median = np.nanmedian(data['est_%s_emissions_std' % tracer].values[valids])
            minimum = np.nanmin(data['est_%s_emissions_std' % tracer].values[valids])
            maximum = np.nanmin(data['est_%s_emissions_std' % tracer].values[valids])
    else:
        median = np.nan
        minimum = np.nan
        maximum = np.nan
        
    nmedian = 100 * median / true_annual_mean
    
    if not quiet:
        print('MB: %.1f (%.0f%%), SD: %.1f (%.0f%%), RMSD: %.1f (%.0f%%), N: %d, Median; %.1f (%.0f%%)' %
              (mb, nmb, sd, nsd, rmsd, nrmsd, n, median, nmedian))
        
    return {
        'bias': float(mb),           'bias [%]': float(nmb),
        'scatter': float(sd),        'scatter [%]': float(nsd),
        'pr': float(pr),             'pr [%]': float(npr),
        'RMSD': float(rmsd),         'RMSD [%]': float(nrmsd),
        'median_std': float(median), 'median_std [%]': float(nmedian),
        'min_std': float(minimum),   'max_std': float(maximum),
        'number_of_cases': float(n),
        'true_annual_mean': float(true_annual_mean)
    }


def fit_ratio(ratio, ratio_std, use_sigma=False):
    """
    Fit a constant CO2:NOX ratio and its uncertainty
    """  
    xdata = np.arange(ratio.size)
    
    if use_sigma:
        sigma = ratio_std.values
        absolute_sigma = True
        ydata = ratio
        valids = np.isfinite(ydata) & np.isfinite(sigma)       
    else:
        sigma = None
        absolute_sigma = False
        ydata = ratio
        
        valids = np.isfinite(ydata)    
        
    if valids.sum() == 0:
        return np.nan, np.nan

    # constant function
    constant = lambda xdata, p: np.full(xdata.shape, p)
    
    p, cov = scipy.optimize.curve_fit(constant, xdata[valids], ydata[valids],
                                      sigma=sigma, absolute_sigma=absolute_sigma)
    
    if use_sigma:
        p_std = np.sqrt(cov[0,0])
    else:
        # compute uncertainty assuming moderate quality of fit
        residual = constant(xdata[valids], p) - ydata[valids]

        chi2 = np.sum(residual**2) / (ydata.size - p.size)
        p_std = np.sqrt(chi2)

    return p[0], p_std


def prepare_ts(data, nsat, n_knots=2):
    """
    Prepare time series for constellation incl. computing seasonal cycle and
    annual emissions, and comparing with true emissions at overpass.
    """
    lon_eqs_in_constellation = smartcarb.misc.lon_eq_by_nsat(nsat)
    mask = data.lon_eq.isin(lon_eqs_in_constellation)
    
    # create new dataset for given constellation
    data = data.where(mask, drop=True)
    data.attrs['constellation'] = nsat
    
    # fix that where adds extra dimension
    for name in data:
        if 'time' in data[name].dims:
            data[name] = data[name].mean('overpass')
    
    # fit cycles
    ratios = data['est_CO2:NO2_emissions'].values.copy()
   
    # mask outliers
    Q1 = scipy.stats.scoreatpercentile(ratios[np.isfinite(ratios)], 25)
    Q2 = scipy.stats.scoreatpercentile(ratios[np.isfinite(ratios)], 50)
    Q3 = scipy.stats.scoreatpercentile(ratios[np.isfinite(ratios)], 75)
    mask = (ratios < (Q1 - 1.5 * (Q3 - Q1))) | (ratios > (Q3 + 1.5 * (Q3 - Q1)))
    
    
    ratios[mask] = np.nan
    data['est_CO2:NO2_emissions'].values[:] = ratios    
   
    ratio, ratio_std = fit_ratio(ratios, None, use_sigma=False)
    data['est_CO2:NO2_emissions'].attrs['mean_ratio'] = ratio
    data['est_CO2:NO2_emissions'].attrs['mean_ratio_std'] = ratio_std
    
    stats = compute_statistics(data, 'CO2:NO2', lon_eqs_in_constellation,
                               true_annual_mean=data['true_CO2:NO2_emissions'].mean(),
                               quiet=True
                              )
    data['est_CO2:NO2_emissions'].attrs.update(stats)
    
    for trace_gas in ['CO2', 'NO2']:
        tp.add_cycle(data, n_knots=n_knots, nsat=nsat, gamma=None,
                     varname='est_%s_emissions' % trace_gas,
                     cyclename='est_%s_cycle' % trace_gas,
                     use_error=True)    
        
        n_overpass = data[f'est_{trace_gas}_cycle'].attrs['number of overpasses']
        
        # Hill and Nassar (for US power plants)
        if n_overpass == 0:
            data[f'est_{trace_gas}_cycle'].attrs['relative_time_profile_std'] = np.nan
        else:
            data[f'est_{trace_gas}_cycle'].attrs['relative_time_profile_std'] = 100 * np.sqrt(0.286784**2 / n_overpass + 0.3100965 / n_overpass)
        
        true_annual_emissions = data['true_%s_emissions' % trace_gas].mean()
        stats = compute_statistics(data, trace_gas, lon_eqs_in_constellation,
                                   true_annual_mean=true_annual_emissions,
                                   quiet=True)

        data['est_%s_emissions' % trace_gas].attrs.update(stats)
        
    return data



def plot_ts(data, n_knots=2, show_errors=False, fig=None, axes=None):

    
    lon_eqs_in_constellation = sorted(set(data['lon_eq'].values.astype(int)))
    
    if fig is None or axes is None:
        fig, axes = plt.subplots(3,1, sharex=True, figsize=(6.97, 6.97))


    for ax, tracer in zip(axes, ['CO2', 'NO2', 'CO2:NO2']):
        
        if tracer == 'NO2':
            units = 'kt yr$^{-1}$'
        elif tracer in ['CO2', 'CO2b']:
            units = 'Mt yr$^{-1}$'
        else:
            units = 'ratio'
            
        if tracer == 'CO2:NO2':
            ax.plot(data['time'], data['true_CO2:NO2_emissions'], 'k-',
                    label='True ratio: $R_t$ = %.0f' % np.mean(data['true_CO2:NO2_emissions']))
        else:
            em_date, em_range = vis.extract_time_range(data['time'],
                                                       data['true_%s_emissions' % tracer], 10, 11)
            ax.plot(em_date, em_range.mean(0), 'k-',
                    label='True emissions (10:30 UTC): %.1f %s' % (em_range.mean(), units))
            
            
            #print(em_range.mean(), data['true_%s_emissions' % tracer])

            
        if tracer in ['CO2']:
            ax.set_ylabel('CO$_2$ [%s]' % units)
        elif tracer == 'NO2':
            ax.set_ylabel('NO$_x$ [%s]' % units)
        elif tracer == 'CO2b':
            ax.set_ylabel('NO$_x$ $\cdot$ $R_e$ [%s]' % units)   
        else:
            ax.set_ylabel('CO$_2$:NO$_x$')

            
        for i, lon_eq in enumerate(lon_eqs_in_constellation, 1):

            letter = LON_EQ_TO_LETTER[lon_eq]

            sat = data['lon_eq'].astype(int) == lon_eq

            times = data['overpass'].loc[sat].values
            values = data['est_%s_emissions' % tracer].loc[sat].values

            # use or not use uncertainty estimated from fitted cycle
            if tracer == 'CO2:NO2':
                errors = data['est_CO2:NO2_emissions'].attrs['mean_ratio_std']
            else:
                errors = data['est_%s_emissions_std' % tracer].loc[sat].values

            n_sats = np.sum(np.isfinite(values))
            
            marker = 'osvd^><'[i]

            if show_errors:
                ax.errorbar(times, values, errors, ls='', marker=marker,
                            label='#%s (%d plumes)' % (letter, n_sats))
            else:
                ax.errorbar(times, values, ls='', marker=marker,
                            label='#%s (%d plumes)' % (letter, n_sats))

        
        # plot cycle
        if tracer == 'CO2:NO2':
            
            ratio = data['est_CO2:NO2_emissions'].attrs['mean_ratio']
            ratio_std = data['est_CO2:NO2_emissions'].attrs['mean_ratio_std']            
            
            ax.hlines([ratio], data['time'][0], data['time'][-1], colors='r',
                      label='Mean ratio: $R_e$ = %.0f$\pm$%.0f' % (ratio, ratio_std))
        else:
            unc = data['est_%s_cycle' % tracer].attrs['annual_mean_std']
           
            # time profile uncertainity
            time_std = data['est_%s_cycle' % tracer].attrs.get('relative_time_profile_std', 0.0)
            time_std = time_std / 100 * em_range.mean() # relative -> absolute error
            unc = np.sqrt(unc**2 + time_std**2)
            
            ax.plot(data['time'], data['est_%s_cycle' % tracer], 'r-',
                    label='Estimated emissions: %.1f$\pm$%.1f %s' %
                    (data['est_%s_cycle' % tracer].attrs['annual_mean'],
                     unc, units))
        
        if tracer == 'CO2:NO2':
            true_ratio = np.mean(data['true_CO2:NO2_emissions'])
            bottom, top = ax.get_ylim()
            
            ax.set_ylim(
                min(bottom, np.floor(true_ratio / 500) * 500 - 1000),
                max(top, np.ceil(true_ratio / 500) * 500 + 1000)
            )

        ax.grid(True)

        ax.xaxis.set_major_formatter(DateFormatter('%b'))
        ax.set_xlim(datetime(2015,1,1), datetime(2015,12,31))

        ax.legend(fontsize='x-small', ncol=3, loc=1)
        
   
    # ylim for CO2 and NO2 emissions
    ymin = min(data['est_CO2_emissions'].min(), data['est_NO2_emissions'].min()) - 5.0
    ymax = max(data['est_CO2_emissions'].max(), data['est_NO2_emissions'].max()) + 5.0
    
    ymin = min(0, np.floor(ymin / 10) * 10)
    ymax = np.ceil(ymax / 10) * 10
        
    axes[0].set_ylim(ymin, ymax)
    axes[1].set_ylim(ymin, ymax)
        
        
    fig.align_ylabels(axes)
    #plt.tight_layout(pad=0.05)
    ax.set_xlabel('2015')
    #plt.tight_layout(pad=0.05)
    
    return data




def plot_time_profiles():
    fig, axes = plt.subplots(1,3, figsize=(1.5*6.97, 1.5*4.31/3))
    tp.plot_diurnal_cycle('Janschwalde', norm=True, ax=axes[0], add_legend=False, only_co2=True)
    tp.plot_weekly_cycle('Janschwalde', norm=True, ax=axes[1], add_legend=False, only_co2=True)
    tp.plot_seasonal_cycle('Janschwalde', norm=True, ax=axes[2], add_legend=False, only_co2=True, short_month=True)

    for i, ax in enumerate(axes):
        ax.set_ylabel('')
        ax.set_ylim(0.6, 1.4)
        ax.set_title('(%s)' % 'abc'[i])


    axes[0].set_ylabel('Time profile')

    plt.tight_layout(pad=0.1)

    plt.savefig(os.path.join(OUTPUT_PATH, 'snap1_time_profiles.png'))
    plt.savefig(os.path.join(OUTPUT_PATH, 'snap1_time_profiles.pdf'))



def get_output_filename(filename, path='detections', extension='_detection.nc'):
    filename = os.path.splitext(os.path.basename(filename))[0]
    return os.path.join(OUTPUT_PATH, path, filename + extension)



def load_dataset(filename):
    
    with open(filename) as nc_file:
        data = xarray.open_dataset(filename).load()
        
    for key in data:
        if data[key].dtype == np.dtype('int8'):
            data[key] = data[key].astype(bool)
    
    return data



def write_to_netcdf(data, filename):
    """
    Write xarray.Dataset to netCDF. To avoid HDF5 errors in existing files,
    write a temporary file and replace existing file afterwards.
    """
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    
    # Write temporary file and close dataset.
    data.to_netcdf(filename + '.tmp')
    data.close()
    
    # Rename filename
    try:
        os.remove(filename)
    except FileNotFoundError:
        pass
    shutil.move(filename + '.tmp', filename)


    
def read_curves(filename, sources=None):
    """
    Read curves from netCDF file.
    """
    if sources is None:
        with netCDF4.Dataset(filename) as d:
            sources = list(d.groups.keys())
    
    curves = {}
        
    for source in sources:
        try:
            curves[source] = gauss.Poly2D.from_file(filename, group=source)
        except OSError as e:
            print(e.errno)
            
    return curves
    
    
def write_curves_to_netcdf(curves, filename):
    for source, curve in curves.items():
        if curve is not None:
            curve.to_file(filename=filename, group=source)
    

    
def select_source(data, source):
    """
    Select source in data but also compute new fields for other plumes and multiple sources.
    """
    this = data.sel(source=source)
    this['other_sources'] = data['detected_plume'].any('source') & ~this['detected_plume']
    this['multiple_sources'] = data['detected_plume'].sum('source') > 1
    
    return this



    
def get_mass_balance_data(fn, sources, make_figure=False, redo_detection=False, 
                          tracer='NO2', scaling=1.0, ssp='high', quiet=False):
    """
    
    """
    
    if tracer == 'NO2':
        co2_ssp = 0.7
        co2_scaling = 1.0 # scaling always applies to NO2 
        
        no2_ssp = ssp
        no2_scaling = scaling
        
    elif tracer == 'CO2':
        co2_ssp = {'low': 0.5, 'medium': 0.7, 'high': 1.0}[ssp]
        co2_scaling = 1.0
        
        no2_ssp = 'high'
        no2_scaling = scaling
    
    
    path = os.path.join('%s-%s-scaling-%.1f' % (tracer, ssp, scaling), 'detections')
    output_fname = get_output_filename(fn, path=path, extension='_detection.nc')
    curves_fname = get_output_filename(fn, path=path, extension='_curves.nc')
    os.makedirs(os.path.dirname(output_fname), exist_ok=True)
    
   
    # detect plume and add center line
    if redo_detection:
        if not quiet:
            print('Detect plumes')
            
        data = smartcarb.io.read_level2(fn, co2_ssp, 0.01, no2_ssp, 0.3,
                                        seed='orbit', use_constant=False, add_berlin=False,
                                        only_observations=True, co2_scaling=co2_scaling,
                                        no2_scaling=no2_scaling)

        data = dplume.detect_plumes(data, sources, variable=tracer, variable_std=f'{tracer}_std',
                                    ns=0.5, local_filter='gaussian')

        data.attrs.update(
                    DESCRIPTION = ('Synthetic XCO2 and NO2 satellite image with ',
                                   'auxiliary data for estimating CO2/NOx ',
                                   'emissions of cities and point sources'),
                    DATAORIGIN = 'SMARTCARB study',
                    CREATOR = 'Gerrit Kuhlmann',
                    EMAIL = 'gerrit.kuhlmann@empa.ch',
                    AFFILIATION = 'Empa Duebendorf, Switzerland',
                    DATE_CREATED = pandas.Timestamp.now().strftime('%Y-%m-%d %H:%M'),
        )

        data, curves = dplume.describe_plume(data, variables=['CO2', 'NO2'], sources=None, plume_area=25e3, force_origin=True)
            
        if os.path.exists(output_fname):
            os.remove(output_fname)
        data.to_netcdf(output_fname)
        
        if os.path.exists(curves_fname):
            os.remove(curves_fname)
        write_curves_to_netcdf(curves, curves_fname)

        
    else:
        try:
            data = load_dataset(output_fname)
            if not quiet:
                print('Load:', output_fname)
        except FileNotFoundError:
            data = None
            if not quiet:
                print('File not found:', output_fname)      
            
        try:
            curves = read_curves(curves_fname)
            if not quiet:
                print('Read:', curves_fname)      
        except FileNotFoundError:
            curves = None
            if not quiet:
                print('File not found:', curves_fname)      
    
    if make_figure:
        for tracer in ['CO2', 'NO2']:
            domain = smartcarb.config.DOMAINS['smartcarb']
            fig = dplume.vis.show_detected_plumes(data, curves, add_polygons=True,
                                                  domain=domain, variable=tracer)

            for ext in ['png', 'pdf']:
                figure_fname = get_output_filename(fn, path=path, extension='_%s_with_detections.%s' % (tracer, ext))
                fig.savefig(figure_fname)
    
    return data, curves



def show_setup(data, curves, ld=None, add_polygons=True, tracer='CO2', zoom_on=None, figwidth=6.97,
               ax=None, cax=None, cmap=None, alpha=None, min_height=None, add_multiple_sources=True,
               vmin=None, vmax=None, add_wind=True):
    """
    Show detected plumes in swath and add curves and control polygons if available.
    """
    first = True # first time adding polygons
    
    fig = dplume.vis.visualize(data, do_zoom=True, show_clouds=True, legend_loc='upper left',
                               tracer=tracer, zoom_on=zoom_on, figwidth=figwidth, ax=ax, cax=cax,
                               cmap=cmap, alpha=alpha, min_height=min_height, vmin=vmin, vmax=vmax,
                               add_multiple_sources=add_multiple_sources, add_wind=add_wind)
    
    if ax is None:
        ax, cax = fig.get_axes()
        
        
    for i, source in enumerate(data.source.values):
    
        if zoom_on is not None and source != zoom_on:
            continue
    
        d = data.sel(source=source)

        if source in curves and curves[source] is not None:
            
            plume = d['detected_plume']
        
            if ld is None:
                xa, xb, ya, yb = compute_polygons(d, delta=5e3,
                                                  add_upstream_box=True)
            else:
                xa, xb = ld[i].xa, ld[i].xb
                ya, yb = ld[i].ya, ld[i].yb
                
            
            # don't draw curves for overlapping plumes
            if misc.has_multiple_sources(data, source):
                continue
                
            x = d['x'].values[plume]
            y = d['y'].values[plume]
            curve = curves[source]
         
            tmax = curve.arc2parameter(xb[-1])
            t = np.linspace(curve.t_o, tmax)            
            
            smartcarb.map.add_curve(ax, *curve(t=t), zorder=10,
                                    color='black', label='Center lines' if first else None)
            
            if add_polygons:
                
                # polygon upstream
                mass.add_area(ax, d, curve, xa[:1], xb[:1], ya[:1], yb[:1],
                              add_label=first, lw=1,
                              do_middle_lines=True)     
                
                # polygons downstream
                mass.add_area(ax, d, curve, xa[1:], xb[1:], ya[1:], yb[1:],
                              add_label=False, lw=1,
                              do_middle_lines=True)     
                
            
            smartcarb.map.add_curve(ax, curve.x0, curve.y0, marker='D',
                                    ls='', color='yellow', mec='k', ms=5,
                                    label='Origin ($x_o$, $y_o$)' if first else None)
            first = False
            
    return fig



def plot_mass_balance(ld, time, origin, figwidth=6.97, ax1=None, ax2=None, no2_scaling=1.0):

    if ax1 is None or ax2 is None:
        fig, (ax1, ax2) = plt.subplots(2,1, sharex=True, figsize=(6.97,4))
    else:
        fig = ax1.get_figure()

    factor = gauss.kgs_to_Mtyr(1.0)
    
    ax1.errorbar(ld['along'] / 1e3, factor * ld['CO2_flux'], factor * ld['CO2_flux_std'],
                 marker='o', ms=4, ls='', label='Along-plume estimates')
    ax2.errorbar(ld['along'] / 1e3, 1e3 * factor * ld['NO2_flux'], 1e3 * factor * ld['NO2_flux_std'],
                 marker='o', ms=4, ls='', label='Along-plume estimates')

    
    # CO2
    Q = ld['CO2_flux_fit'].attrs['emissions']
    u = ld['CO2_flux_fit'].attrs['wind_speed']
    Q_std = ld['CO2_flux_fit'].attrs['emissions_std']

    x = np.linspace(1.0, ld['along'][-1] + 2.5e3, 501)
    y = Q * np.ones_like(x)
    
    label = 'Estimated emissions (Q = %.1f$\pm$%.1f Mt yr$^{-1}$, u = %.1f m s$^{-1}$)' % (Q, Q_std, u)
    label = 'Estimated emissions'
    ax1.plot(x, y, label=label, color='C0', ls='--')

    # NO2
    Q = ld['NO2_flux_fit'].attrs['emissions']
    u = ld['NO2_flux_fit'].attrs['wind_speed']
    tau = ld['NO2_flux_fit'].attrs['decay_time']
    
    Q_std = ld['NO2_flux_fit'].attrs['emissions_std']
    tau_std = ld['NO2_flux_fit'].attrs['decay_time_std']
   
    if np.isnan(tau):
        y = np.full(x.shape, Q)
    else:
        y = gauss.point_plume_model(x, Q, u * tau * 3600)
        
    label = 'Estimated emissions (Q = %.1f$\pm$%.1f kt yr$^{-1}$, u = %.1f m s$^{-1}$, $\\tau$ = %.1f$\pm$%.1f h)' % (Q, Q_std, u, tau, tau_std)    
    label = 'Estimated emissions' 
    ax2.plot(x / 1e3, y, label=label, color='C0', ls='--')

    # true emissions
    q_CO2 = gauss.read_true_emissions(time, tracer='CO2', origin=origin, tmin=10, tmax=11).mean()
    q_NOx = no2_scaling * gauss.read_true_emissions(time, tracer='NO2', origin=origin, tmin=10, tmax=11).mean()   
    
    right = ld['along'][-1] / 1e3 + 2.5
    ax1.hlines([q_CO2], 0.0, right, color='C1', label='True emissions')# (Q = %.1f Mt yr$^{-1}$)' % q_CO2)
    ax2.hlines([q_NOx], 0.0, right, color='C1', label='True emissions')# (Q = %.1f kt yr$^{-1}$)' % q_NOx)

    # labels
    ax1.set_ylabel('CO$_2$ flux [Mt yr$^{-1}$]')
    ax2.set_ylabel('NO$_\mathrm{x}$ flux [kt yr$^{-1}$]')
    
    for ax in [ax1, ax2]:
        if q_NOx < 25:
            ax.set_ylim(-10, 60)        
        else:
            ax.set_ylim(-25, 100)
        ax.grid(True)

        ax.legend(fontsize='small', ncol=1)

        right = ld['along'].max() / 1e3 + 5.0
        xticks = np.array( sorted(set(list(ld.xa.values) + list(ld.xb.values))) ) / 1e3
        ax.set_xticks(xticks)
        ax.set_xlim(xticks[0], xticks[-1])
        
    ax2.set_xlabel('Along plume distance [km]')
    

    
    return fig



def add_lines(ax, edgecolor='k'):
    ax.coastlines(resolution='10m', color=edgecolor, linewidth=1.0)
    lines = cfeature.NaturalEarthFeature(
        category='cultural',
        name='admin_0_boundary_lines_land',
        scale='10m',
    )
    ax.add_feature(lines, edgecolor=edgecolor, facecolor='none', linewidth=1.0)

    lines = cfeature.NaturalEarthFeature(
            category='cultural',
            name='admin_1_states_provinces_lines',
            scale='10m',
    )
    ax.add_feature(lines, edgecolor=edgecolor, facecolor='none', linewidth=0.5)
    

def compute_plume_age_and_length(ld):
    """
    Estimate plume age and length based on wind speed and arc length of
    detected pixels.
    """
    values = ld.x.values[ld.is_plume.values]

    if np.size(values) > 0:
        plume_length = ld.x.values[ld.is_plume.values].max() / 1e3
    else:
        plume_length = 0.0
        
    plume_age = plume_length / (ld.CO2_flux_fit.attrs['wind_speed'] * 3600 / 1000)
    
    return plume_age, plume_length

    
def add_text(ax, data, ld, source):
    time = pandas.Timestamp(data.time.values)
    
    this = data.sel(source=source)
    plume_age, plume_length = compute_plume_age_and_length(ld)

    text = """\
        {satellite} ({time}, Orbit: {orbit})

        Instrument scenario:
        • CO$_2$: $\sigma_\\mathrm{{VEG50}}$ = 0.7 ppm
        • NO$_2$: $\sigma_\\mathrm{{ref}}$ = 2$\\times$10$^{{15}}$ molcules cm$^{{-2}}$

        Plume information:
        • Detection: {detection_tracer} (q = {q}, $\sigma_g$ = {ns} px)
        • Plume size: {plume_size:d} px
        • Plume length: {plume_length:.1f} km
        • Estimated plume age: {plume_age:.1f} h
        • Mean wind: {speed:.1f} m s$^{{-1}}$, {direction:.0f} degrees
        • Angle between curve and wind: {angle:.0f} degrees

        True emissions (10:30 UTC):
        • CO$_2$: {true_CO2_flux:.1f} Mt yr$^{{-1}}$
        • NO$_x$: {true_NO2_flux:.1f} kt yr$^{{-1}}$

        Estimated emissions:
        • CO$_2$ flux: {CO2_flux:.1f}$\pm${CO2_flux_std:.1f} Mt yr$^{{-1}}$
        • NO$_x$ flux: {NO2_flux:.1f}$\pm${NO2_flux_std:.1f} kt yr$^{{-1}}$
        • NO$_x$ decay time: {NOX_decay:.1f}$\pm${NOX_decay_std:.1f} h


    """.format(satellite='CO2M%s' % vis.LON_EQ_TO_LETTER.get(data.lon_eq, ''),
               time=time.strftime('%d %b %Y, %H:%M UTC'),
               orbit=data.orbit,
               detection_tracer=data.attrs['tracer'].replace('2', '$_2$'),
               q=data.attrs['probability for z-value'],
               ns=data.attrs['size of neighborhood'],
               speed=ld.CO2_flux_fit.attrs['wind_speed'],
               direction=ld.CO2_flux_fit.attrs['wind_direction'],
               angle=ld.attrs['angle_between_curve_and_wind'],
               CO2_flux=ld.CO2_flux_fit.attrs['emissions'],           
               CO2_flux_std=ld.CO2_flux_fit.attrs['emissions_std'],  
               NO2_flux=ld.NO2_flux_fit.attrs['emissions'],           
               NO2_flux_std=ld.NO2_flux_fit.attrs['emissions_std'],       
               NOX_decay=ld.NO2_flux_fit.attrs['decay_time'],           
               NOX_decay_std=ld.NO2_flux_fit.attrs['decay_time_std'],
               plume_length=plume_length,
               plume_age=plume_age,
               plume_size=int(this.detected_plume.values.sum()),
               true_CO2_flux=gauss.read_true_emissions(time, 'CO2', source, tmin=10, tmax=11).mean(),
               true_NO2_flux=gauss.read_true_emissions(time, 'NO2', source, tmin=10, tmax=11).mean()
              )
    text = textwrap.dedent(text)
    plt.figtext(0.76, 0.47, text, va='top', linespacing=1.1)

    print(gauss.read_true_emissions(time, 'CO2', source, tmin=10, tmax=11).mean())
    print(gauss.read_true_emissions(time, 'NO2', source, tmin=10, tmax=11).mean())
    
    

def plot_source_overview(data, curves, line_densities, source, no2_scaling=1.0):
    domain = smartcarb.cfg.DOMAINS['smartcarb']
    time = pandas.Timestamp(data.time.values)

    with plt.style.context({'font.size': 10}):
        # figure size (13.94, 6.5) fits on wide screen power point and two figures per page in Copernicus journal.
        scaling = 1.0
        figsize= (13.94*scaling, 7.45 * scaling)
        fig = plt.figure(figsize=figsize)

        # size of level-2 image
        aspect = (smartcarb.cfg.rlon[-1] - smartcarb.cfg.rlon[0]) / (smartcarb.cfg.rlat[-1] - smartcarb.cfg.rlat[0])
        height = (figsize[1] - 0.85) / 2
        width =  height * aspect
        dy = 0.85 / 4 / figsize[1]
        rel_height = height / figsize[1]
        rel_width = width / figsize[0]

        axes = [None, None]
        caxes = [None, None]

        for i in [0,1]:
            axes[i] = fig.add_axes([
                0.035,
                dy + i * (rel_height + 2 * dy) - 0.005,
                rel_width,
                rel_height],
                projection=domain.proj
            )
            axes[i].set_aspect('equal', adjustable='box')
            axes[i].set_xlim(domain.startlon, domain.stoplon)
            axes[i].set_ylim(domain.startlat, domain.stoplat)
            add_lines(axes[i])

            axes[i].set_title('(%s)' % 'ba'[i])

            caxes[i] = fig.add_axes([
                0.04 + rel_width,
                dy + i * (rel_height + 2 * dy) - 0.005,
                0.01, 
                rel_height]
            )

            domain = smartcarb.config.DOMAINS['smartcarb']
            fig = dplume.vis.show_detected_plumes(data, curves, add_polygons=True,
                                                  domain=domain, variable=['NO2', 'CO2'][i],
                                                  zoom_on=source, ax=axes[i], cax=caxes[i])

        # mass balance
        h = (dy + rel_height - 0.06 - 0.015) / 2
        left = 0.02 + rel_width + 0.12

        # other axes
        ax1 = fig.add_axes([
            left,
            0.06 + 0.015 + h,
            0.32,  #0.38,
            h,
        ])
        ax2 = fig.add_axes([
            left,
            0.06,
            0.32, #0.38,
            h])
        
    
        plot_mass_balance(line_densities, time=time, origin=source, ax1=ax1,
                          ax2=ax2, no2_scaling=no2_scaling)       
        ax1.set_xticklabels([])
        ax1.set_title('(d)')


        w = 0.02 + rel_width + 0.12 + 0.38 + 0.02
        
        #tax = fig.add_axes([
        #    w,
        #    0.01,
        #    1.0 - w - 0.02,     
        #    2*h + 0.015 + 0.05,
        #])
        #
        #tax.set_xticklabels([])
        #tax.set_yticklabels([])
        add_text(None, data, line_densities, source)
        
        if line_densities.pixels.size == 0:
            return fig
        
        
        # cross sections for first six cases (  0.98)
        axes = []
        
        # upper row
        for i in [0,1,2]:
            ax = fig.add_axes([
                left + i * (0.58/3 - 0.01 + 0.01),
                dy + (rel_height + 2 * dy) + 0.05 + h,
                0.58/3 - 0.01,
                h - 0.01,
            ])
            
            axes.append(ax)
        
        # lower row
        for i in [0,1,2]:            
            ax = fig.add_axes([
                left + i * (0.58/3 - 0.01 + 0.01),
                dy + (rel_height + 2 * dy) + 0.04,
                0.58/3 - 0.01,
                h - 0.01,
            ])
            axes.append(ax)
        
        # plot line densities
        n = line_densities.along.size
        
        if n <= 6:
            alongs = np.arange(n)
        else:
            alongs = [0, 1, 2, 3, n-2, n-1]
        
        for i, ax in zip(alongs, axes):
            
            if 0 <= i < line_densities.along.size:
                polygon = line_densities.isel(along=i)
                
                with warnings.catch_warnings():
                    warnings.filterwarnings('ignore', r'All-NaN (slice|axis) encountered')
                    
                    vis.plot_across_section(polygon, tracer='both',
                                            method=polygon.method,
                                            show_true=False,
                                            add_errors=True, ax=ax, no_twinx=True,
                                            legend='simple')

                ax.text(0.01, 0.98, '%d-%d km' % (polygon.xa/1e3, polygon.xb/1e3),
                        va='top', ha='left', transform=ax.transAxes)
                
                ax.set_ylabel('Column densities')
            
        bottom = min(ax.get_ylim()[0] for ax in axes)
        top = max(ax.get_ylim()[1] for ax in axes)
        
        # update labels etc.
        for i, ax in enumerate(axes):
            ax.set_ylim(bottom, top)
            
            if i == 1:
                ax.set_title('(c)')
            
            if i in [0,1,2]:
                ax.set_xticklabels([])
                ax.set_xlabel('')
            
            if i not in [0,3]:
                ax.set_yticklabels([])
                ax.set_ylabel('')
                
            if i != 5:
                legend = ax.get_legend()
                
                if legend is not None:
                    legend.remove()
            else:
                ax.legend(ncol=4, loc='lower right')
        
        return fig
    
    


def compute_polygons(data, delta=5e3, add_upstream_box=True):
    """
    Compute [xa,xb] and [ya,yb] intervals for polygons.
    """
    dmax = np.nanmax(data.xp.values[data.detected_plume])
    distances = np.arange(0.0, dmax+delta, delta)
   
    if add_upstream_box:
        xa_values = np.concatenate([[-12e3], distances[:-1]])
        xb_values = np.concatenate([[-2e3], distances[1:]])
    else:
        xa_values = distances[:-1]
        xb_values = distances[1:]
    
    ya, yb = mass.get_plume_width(data, dy=2e3)
    
    return xa_values, xb_values, np.full_like(xa_values, ya), np.full_like(xb_values, yb)
    
    

def compute_emissions_and_uncertainties(tracer, along, line_density, line_density_std, wind, wind_std):
    """\
    Compute emissions (q in kg/s) and for NO2 decay times (tau in hours) as well as their uncertainties.
    
    tracer             CO2 or NO2
    along              along-plume coords in meters
    line_density       CO2 or NO2 line densities
    line_density_std   uncertainty of line densities
    wind               estimate of wind speed in plume
    wind_std           uncertainty of wind speed
    
    """
    # compute flux and uncertainty (not including wind_std yet)
    flux = wind * line_density
    flux_std = wind * line_density_std
    
    _, _, fit, p, p_std = gauss.fit_emissions(tracer, along, flux, flux_std, dmin_fit=0.0, dmax_fit=np.inf, is_point_source=True)
    
    # only LD uncertainty
    q = p[0]
    q_std = np.sqrt(p_std[0]**2 + (p[0] / wind)**2 * wind_std**2 )
       
    # decay time
    if tracer == 'NO2':
        tau = p[1] / wind / 3600
        tau_std = np.sqrt(p_std[1]**2 / wind**2 + p[1]**2 / wind**4 * wind_std**2) / 3600
    else:
        tau = np.nan
        tau_std = np.nan
        
    return q, q_std, tau, tau_std, fit


def estimate_fluxes(data, curve, method='two-gauss',
                    origin=None, show=False, delta=5e3,
                    wind_speed=None, wind_direction=None):
    """
    Estimate fluxes in along-plume direction.
    """
    time = pandas.Timestamp(data['time'].values)
    
    if wind_speed is None or wind_direction is None:
        wind_info = gauss.get_wind_speed(time, hmin=0.0, hmax=1000.0, origin=origin)
        wind_speed = wind_info['mean wind speed (m/s)']
        wind_direction = wind_info['mean wind direction (deg)']

    if np.any(np.isnan(data.xp.values[data.detected_plume])):
        return None
           
    # add one polygon [-10 km - -2 km] upstream for checking for upstream sources
    xa_values, xb_values, ya, yb = compute_polygons(data, delta=delta, add_upstream_box=True)
    
    results = []
    
    for xa, xb in zip(xa_values, xb_values):
            
        if method == 'two-gauss':
            
            # use "sub-areas" upstream where no plume should exists 
            this_method = 'two-gauss' if xa >= 0.0 else 'sub-areas'
                
            ld = mass.extract_line_density(data, 'CO2_minus_estimated_background_mass', 
                                           xa=xa, xb=xb, ya=ya, yb=yb, show=show, show_true=False,
                                           method=this_method, share_mu=True)
            results.append(ld)
            
            
        else:
            ld = mass.extract_line_density(data, 'CO2_minus_estimated_background_mass', 
                                           xa=xa, xb=xb, show=False, show_true=False,
                                           method=method)
            results.append(ld)
            
    if len(results) == 0:
        return None
   

    results = mass.concat_polygons(results)
    results['along'] = xarray.DataArray(0.5 * (results.xa + results.xb), dims='along')
    
    for tracer in ['CO2', 'NO2']: # TODO
        
        if '%s_line_density' % tracer not in results:
            continue
       
        wind_speed_std = 0.5
        
        q, q_std, tau, tau_std, fit = compute_emissions_and_uncertainties(tracer, 
                                                                          results['along'].values,
                                                                          results[f'{tracer}_line_density'].values,
                                                                          results[f'{tracer}_line_density_std'].values,
                                                                          wind_speed,
                                                                          wind_speed_std
                                                                         )
        
        # compute flux and its stds adding wind speed std
        results[f'{tracer}_flux'] = wind_speed * results[f'{tracer}_line_density']
        results[f'{tracer}_flux_std'] = np.sqrt(
            wind_speed**2 * results[f'{tracer}_line_density_std']**2 + 
            wind_speed_std**2 * results[f'{tracer}_line_density']**2
        )
        
        # compute uncertainty of emissions including wind speed
        scaling = 1.0 if tracer == 'CO2' else 1e3
        q = scaling * gauss.kgs_to_Mtyr(q)
        q_std = scaling * gauss.kgs_to_Mtyr(q_std)
        
    
        try:
            angle = gauss.compute_angle_between_curve_and_wind(origin, data, curve, hmin=0.0, hmax=1e3)
            results.attrs['angle_between_curve_and_wind'] = angle
        except FileNotFoundError as e:
            print(e)
        except KeyError as e:
            print(e)
        
        results['%s_flux_fit' % tracer] = xarray.DataArray(fit, dims='along')
        results['%s_flux_fit' % tracer].attrs['emissions'] = q
        results['%s_flux_fit' % tracer].attrs['emissions_std'] = q_std
        results['%s_flux_fit' % tracer].attrs['wind_speed'] = wind_speed
        results['%s_flux_fit' % tracer].attrs['wind_direction'] = wind_direction
        
        if tracer == 'NO2':
            results['%s_flux_fit' % tracer].attrs['decay_time'] = tau
            results['%s_flux_fit' % tracer].attrs['decay_time_std'] = tau_std
            
    
    
    return results
   
    
def detect_plumes(filename, tracer, ssp, scaling):
    
    # get sources (TODO: add to function)
    sources = sorted(smartcarb.cfg.POINT_SOURCES)
    lon_o, lat_o, radius = smartcarb.misc.get_sources(sources, threshold=4.0)
    sources = xarray.Dataset(coords={'source': sources})
    sources['lon_o'] = xarray.DataArray(lon_o, dims='source')
    sources['lat_o'] = xarray.DataArray(lat_o, dims='source')
    sources['radius'] = xarray.DataArray(radius, dims='source')
    
    data, curves = get_mass_balance_data(filename, sources, make_figure=True, redo_detection=True,
                                         tracer=tracer, ssp=ssp, scaling=scaling)
    
    return data, curves
    
    
def detect_plumes_mp(args):
    detect_plumes(filename=args[0], **args[1])
    plt.close('all')
    
    

def estimate_emissions(source, filename=None, make_figure=True, tracer='NO2',
                       ssp='high', scaling=1.0, without_no2=False):
    """
    Estimate emissions of point source.
    """
    
    # get sources (TODO: add to function)
    sources = sorted(smartcarb.cfg.POINT_SOURCES)
    lon_o, lat_o, radius = smartcarb.misc.get_sources(sources, threshold=4.0)
    sources = xarray.Dataset(coords={'source': sources})
    sources['lon_o'] = xarray.DataArray(lon_o, dims='source')
    sources['lat_o'] = xarray.DataArray(lat_o, dims='source')
    sources['radius'] = xarray.DataArray(radius, dims='source')
    
    data, curves = get_mass_balance_data(filename, sources, make_figure=False, redo_detection=False,
                                         tracer=tracer, ssp=ssp, scaling=scaling)

    if data is None or curves is None:
        print('Data or curves file does not exists (%s)' % filename)
        return None
    
    if source not in curves or source not in data.source:
        print('"%s" not in swath or not detected' % source)
        return None
    
    
    this = select_source(data, source)
    
    if this['detected_plume'].sum() == 0 or misc.has_multiple_sources(data, source):
        print('No detection or multiple sources for %s' % source)
        return None
    
    if curves.get(source, None) is None:
        print('No curve for %s' % source)
        return None
    
    
    if without_no2:
        this['NO2_minus_estimated_background_mass'].values[:] = np.nan
                
                
    flux = estimate_fluxes(this, curves[source], origin=source, method='two-gauss')
   
    if flux is None:
        print('Failed to compute any line densities for %s.' % source)
        return None
        
    if without_no2:
        path = os.path.join('%s-%s-scaling-%.1f-without-no2' % (tracer, ssp, scaling), source)
    else:
        path = os.path.join('%s-%s-scaling-%.1f' % (tracer, ssp, scaling), source)
    os.makedirs(os.path.join(OUTPUT_PATH, path), exist_ok=True)
    
    flux.attrs.update(
        DESCRIPTION = ('Estimated CO2 and NO2/NOx emissions with '
                       'mass-balance approach using detected  plumes.'),
        DATAORIGIN = 'SMARTCARB study',
        CREATOR = 'Gerrit Kuhlmann',
        EMAIL = 'gerrit.kuhlmann@empa.ch',
        AFFILIATION = 'Empa Duebendorf, Switzerland',
        DATE_CREATED = pandas.Timestamp.now().strftime('%Y-%m-%d %H:%M'),
    )
    flux_filename = get_output_filename(filename, path=path, extension='_flux.nc')
    flux.to_netcdf(flux_filename)
    
    if make_figure:
        fig = plot_source_overview(data, curves, flux, source, no2_scaling=scaling)
        
        for ext in ['png', 'pdf']:
            figure_fname = get_output_filename(filename, path=path, extension='_flux.%s' % ext)
            fig.savefig(figure_fname)
    
    return flux
   
    
def estimate_emissions_mp(args):
    estimate_emissions(filename=args[0], **args[1])
    plt.close('all')
  
    
def main(source, tracer='NO2', ssp='high', scaling=1.0, n_proc=1, only_detection=False):  
    
    with open('/project/kug/repos/smartcarb/plume-detection-algorithm/files/orbits-with-cloud-free-plumes.dat') as file_:
        good_filenames = [line.strip() for line in file_ if not line.startswith('#')]

    good_filenames = (sorted(set(good_filenames)))

    # largest power plants with >3.0 kt NO2 yr-1 emission
    origins = sorted(smartcarb.cfg.POINT_SOURCES)
   
    if n_proc == 1:
        for fn in good_filenames:
            if only_detection:
                detect_plumes(fn, tracer, ssp, scaling)
            else:
                estimate_emissions(source, filename=fn, make_figure=True, tracer=tracer,
                                   ssp=ssp, scaling=scaling, only_detection=False)
            plt.close('all')
    else:
        if only_detection:
            iterator = [(fn, {'tracer': tracer, 'ssp': ssp, 'scaling': scaling})
                        for fn in good_filenames]

            with mp.Pool(processes=n_proc) as pool:
                pool.map(detect_plumes_mp, iterator)
            
        else:
            iterator = [(fn, {'source': source, 'make_figure': True, 'tracer': tracer, 'ssp': ssp,
                              'scaling': scaling})
                        for fn in good_filenames]

            with mp.Pool(processes=n_proc) as pool:
                pool.map(estimate_emissions_mp, iterator)

    
    
    